import os
import uuid
from typing import Callable, List, Optional, Union

import numpy as np
from arbol import aprint
from pydantic import BaseModel

from litemind.agent.augmentations.information.information import Information
from litemind.agent.augmentations.vector_db.default_vector_db import (
    DefaultVectorDatabase,
)
from litemind.apis.base_api import BaseApi
from litemind.media.media_base import MediaBase


class QdrantVectorDatabase(DefaultVectorDatabase):
    """
    A vector database implementation using Qdrant.

    This implementation uses the Qdrant vector database to store and retrieve informations.
    It requires the qdrant-client package to be installed.
    """

    def __init__(
        self,
        name: str = "QdrantVectorDB",
        description: Optional[str] = None,
        collection_name: str = "informations",
        embedding_function: Optional[
            Callable[[List[Information]], List[List[float]]]
        ] = None,
        embedding_length: Optional[int] = None,
        url: Optional[str] = None,
        location: Optional[str] = None,
        qdrant_api_key: Optional[str] = None,
        api: BaseApi = None,
    ):
        """
        Initialize the Qdrant vector database.

        Parameters
        ----------
        name: str
            The name of the augmentation.
        description: Optional[str]
            A description of the augmentation.
        collection_name: str
            The name of the collection in Qdrant.
        embedding_function: Optional[Callable[[List[Information]], List[List[float]]]]
            A function that takes a list of informations and returns a list of embedding vectors.
            If not provided, a default embedding function will be used.
        embedding_length: int
            The length of the embedding vectors.
        url: Optional[str]
            The URL of the Qdrant server. If None, will use a local instance.
        location: Optional[str]
            The location of the local Qdrant instance. Only used if url is None.
        qdrant_api_key: Optional[str]
            The API key for the Qdrant server.
        api: BaseApi
            The API to use for embedding. If not provided, uses litemind's default API.

        """
        super().__init__(
            name=name,
            description=description,
            embedding_function=embedding_function,
            embedding_length=embedding_length,
            api=api,
        )

        self.collection_name = collection_name

        try:
            from qdrant_client import QdrantClient
            from qdrant_client.http import models

            if url:
                self.client = QdrantClient(url=url, api_key=qdrant_api_key)
            else:
                if location is None:
                    # Get the user's home folder:
                    home_folder = os.path.expanduser("~")

                    # Qdrant data folder:
                    location = os.path.join(home_folder, ".qdrant_data")

                    # Create the Qdrant data folder:
                    os.makedirs(location, exist_ok=True)

                # Use the local Qdrant instance at the specified location:
                self.client = QdrantClient(path=location)

            # Create collection if it doesn't exist
            collections = self.client.get_collections().collections
            collection_names = [collection.name for collection in collections]

            if collection_name not in collection_names:
                aprint(f"Creating new Qdrant collection: {collection_name}")
                self.client.create_collection(
                    collection_name=collection_name,
                    vectors_config=models.VectorParams(
                        size=self.embedding_length,
                        distance=models.Distance.COSINE,
                    ),
                )

        except ImportError:
            raise ImportError(
                "Qdrant not installed. Please install it with 'pip install qdrant-client'."
            )

    def add_informations(
        self,
        informations: List[Information],
        embeddings: Optional[List[List[float]]] = None,
    ) -> List[str]:
        """
        Add informations to the database.

        Parameters
        ----------
        informations: List[Information]
            The informations to add to the database.
        embeddings: Optional[List[List[float]]]
            The embeddings for the informations. If not provided, they will be computed.

        Returns
        -------
        List[str]
            The IDs of the added informations.
        """
        from qdrant_client.http import models

        # Generate IDs for informations without IDs
        for info in informations:
            if info.id is None:
                info.set_id(str(uuid.uuid4()))

        # Create embeddings if not provided
        if embeddings is None:
            aprint(f"Computing embeddings for {len(informations)} informations")
            embeddings = self.embedding_function(informations)

        # Prepare points for Qdrant
        points = []
        for info, emb in zip(informations, embeddings):
            # Create a point
            payload = {"info": info.to_base64()}
            points.append(
                models.PointStruct(
                    id=info.id,
                    vector=emb,
                    payload=payload,
                )
            )

        # Upsert points to Qdrant
        self.client.upsert(
            collection_name=self.collection_name,
            points=points,
        )

        return [info.id for info in informations]

    def get_information(self, information_id: str) -> Optional[Information]:
        """
        Retrieve an information by its ID.

        Parameters
        ----------
        information_id : str
            The ID of the information to retrieve.

        Returns
        -------
        Optional[Information]
            The information, or None if not found.
        """
        points = self.client.retrieve(
            collection_name=self.collection_name,
            ids=[information_id],
        )

        if not points:
            return None

        point = points[0]
        return Information.from_base64(point.payload["info"])

    def similarity_search(
        self,
        query: Union[str, BaseModel, MediaBase, Information, List[float]],
        k: int = 5,
        threshold: float = 0.0,
    ) -> List[Information]:
        """
        Find similar informations using Qdrant's vector search.

        Parameters
        ----------
        query : Union[str, BaseModel, MediaBase, Information, List[float]]
            The query: a string, Pydantic model, media object, Information,
            or a raw embedding vector.
        k : int
            The maximum number of results to return.
        threshold : float
            Minimum similarity score. Results below this value are excluded.

        Returns
        -------
        List[Information]
            The most similar informations, ordered by descending similarity.
        """

        # If no informations are requested we obey:
        if k <= 0:
            return []

        if isinstance(query, list) and isinstance(query[0], float):

            # If the query is already a vector, we don't need to normalize it:
            query_vector = query

        else:
            # Normalize the query
            query = self._normalize_query(query)

            # Wrap in Information:
            query = Information(query)

            # Get the embedding for the query:
            query_vector = list(self.embedding_function([query])[0])

        # Convert to numpy array:
        query_vector = np.array(query_vector)

        # Normalize query vector
        query_norm = np.linalg.norm(query_vector)
        if query_norm > 0:
            query_vector = query_vector / query_norm

        # Perform similarity search
        search_result = self.client.query_points(
            collection_name=self.collection_name,
            query=query_vector,
            limit=k,
        )

        results = []
        for result in search_result.points:

            # Deserialize the information from the payload
            info = Information.from_base64(result.payload["info"])
            # Information.from_json(result.payload["info"])

            # Set the score (similarity) for the information
            info.score = result.score

            # Check if the score is above the threshold
            if info.score >= threshold:
                # Add the information to the list
                results.append(info)

        return results

    def delete(self, ids: List[str]) -> None:
        """
        Delete informations from the database.

        Parameters
        ----------
        ids: List[str]
            The IDs of the informations to delete.
        """
        self.client.delete(
            collection_name=self.collection_name,
            points_selector=ids,
        )

    def clear(self) -> None:
        """Clear all informations from the database."""
        try:
            # Delete the collection
            self.client.delete_collection(collection_name=self.collection_name)

            # Recreate the collection
            from qdrant_client.http import models

            self.client.create_collection(
                collection_name=self.collection_name,
                vectors_config=models.VectorParams(
                    size=self.embedding_length,
                    distance=models.Distance.COSINE,
                ),
            )
        except Exception as e:
            # If collection doesn't exist, that's fine
            aprint(f"Error clearing collection: {e}")

    # Implement AugmentationBase methods
    def get_relevant_informations(
        self,
        query: Union[str, BaseModel, MediaBase, Information],
        k: int = 5,
        threshold: float = 0.0,
    ) -> List[Information]:
        """
        Get relevant informations for a query via Qdrant similarity search.

        Parameters
        ----------
        query : Union[str, BaseModel, MediaBase, Information]
            The query to retrieve informations for.
        k : int
            The maximum number of informations to retrieve.
        threshold : float
            Minimum relevance score. Results below this value are excluded.

        Returns
        -------
        List[Information]
            A list of relevant informations sorted by similarity.
        """

        # Normalize the query
        query = self._normalize_query(query)

        # Perform similarity search
        return self.similarity_search(query, k, threshold)

    def save(self) -> None:
        """Persist the database state. No-op for Qdrant (auto-persisted)."""
        pass

    def close(self) -> None:
        """Close the connection to the Qdrant database."""
        if hasattr(self, "client") and self.client is not None:
            self.client.close()



import json
from typing import Callable, List, Optional, Sequence

from pydantic import BaseModel

from litemind.agent.augmentations.augmentation_default import AugmentationDefault
from litemind.agent.augmentations.information.information import Information
from litemind.agent.augmentations.vector_db.base_vector_db import BaseVectorDatabase
from litemind.apis.base_api import BaseApi
from litemind.media.types.media_audio import Audio
from litemind.media.types.media_code import Code
from litemind.media.types.media_document import Document
from litemind.media.types.media_image import Image
from litemind.media.types.media_json import Json
from litemind.media.types.media_object import Object
from litemind.media.types.media_table import Table
from litemind.media.types.media_text import Text
from litemind.media.types.media_video import Video
from litemind.utils.uri_utils import is_uri


class DefaultVectorDatabase(AugmentationDefault, BaseVectorDatabase):
    """Abstract base for vector databases with a default embedding function.

    Provides a multi-modal embedding function that dispatches to the
    appropriate API method (text, image, audio, etc.) based on the media
    type of the informations being embedded.
    """

    def __init__(
        self,
        name: Optional[str] = None,
        description: Optional[str] = None,
        embedding_function: Optional[
            Callable[[List[Information]], List[List[float]]]
        ] = None,
        embedding_length: Optional[int] = None,
        api: BaseApi = None,
    ):
        """
        Initialize the vector database with an embedding function.

        Parameters
        ----------
        name : str, optional
            The name of the augmentation. Defaults to the class name.
        description : Optional[str]
            A description of the augmentation.
        embedding_function : Optional[Callable[[List[Information]], List[List[float]]]]
            A function that takes a list of informations and returns
            embedding vectors. If not provided, uses the built-in
            multi-modal embedding function.
        embedding_length : Optional[int]
            The dimensionality of the embedding vectors. If not provided,
            it is auto-detected from a sample embedding.
        api : BaseApi, optional
            The API to use for embedding. If not provided, uses
            ``CombinedApi()`` as the default.
        """
        super().__init__(name=name, description=description)

        # Set the name of the augmentation to the extended class name:
        if name is None:
            name = self.__class__.__name__
        self.name = name

        # Set the embedding function:
        self.embedding_function = embedding_function or self._default_embedding_function

        # If no api is provided, use the default API (only needed for default embedding):
        if api is None and embedding_function is None:
            try:
                from litemind import CombinedApi

                api = CombinedApi()
            except ValueError:
                raise ValueError(
                    "No API provided and litemind's default API is not available. "
                    "Please provide an API."
                )
        self.api = api

        # Call the embedding function to get the embedding length if not loaded from disk
        if embedding_length is None:
            # Make a simple information that contains some text:
            query_info = Information(Text("sample text"))

            sample_embedding = self.embedding_function([query_info])[0]
            embedding_length = len(sample_embedding)

        # Set the embedding length:
        self.embedding_length = embedding_length

    def _default_embedding_function(
        self, informations: Sequence[Information]
    ) -> Sequence[Sequence[float]]:
        """
        Default embedding function using litemind's API.

        Parameters
        ----------
        informations: Sequence[Information]
            List of informations to embed.

        Returns
        -------
        List[List[float]]
            The embedding vectors.
        """

        # If embedding_length does not exist in object, then we use a default value for the embedding length:
        if not hasattr(self, "embedding_length"):
            self.embedding_length = 768

        # First we check if all the informations are of the same type:
        info_types = {info.type for info in informations}
        if len(info_types) > 1:

            # Initialise the list of embeddings:
            embeddings = []

            # If the informations are of different types we must handle them separately:
            for info in informations:
                # Compute the embedding for the information:
                embeddings_for_info = self._raw_embedding_function([info])

                # Add embedding to the list of embeddings:
                embeddings.append(embeddings_for_info[0])

        else:
            # If all the informations are of the same type, we can proceed normally:
            embeddings = self._raw_embedding_function(informations)

        # Return the embeddings:
        return embeddings

    def _raw_embedding_function(
        self, informations: Sequence[Information]
    ) -> List[List[float]]:
        """
        Compute embeddings for a batch of same-type informations.

        Dispatches to the appropriate API embedding method based on
        media type (text, image, audio, video, document, etc.).

        Parameters
        ----------
        informations : Sequence[Information]
            Informations to embed. All must have the same media type.

        Returns
        -------
        List[List[float]]
            The embedding vectors.

        Raises
        ------
        ValueError
            If the list is empty or the media type is unsupported.
        """

        # Check if list is empty:
        if not informations:
            raise ValueError("No informations provided for embedding.")

        # If the information already has an embedding, we use it:
        if all(info.embedding is not None for info in informations):
            embeddings = [info.embedding for info in informations]
        else:

            if all(info.has_type(Text) for info in informations):

                # If the information is of type text, we use the text embedding function:
                texts = [info.content for info in informations]

                # Embed the texts:
                embeddings = self.api.embed_texts(
                    texts=texts, dimensions=self.embedding_length
                )
            elif all(info.has_type(Code) for info in informations):
                # If the information is of type code, we use the code embedding function:
                codes = [info.content for info in informations]

                # Embed the codes:
                embeddings = self.api.embed_texts(
                    texts=codes, dimensions=self.embedding_length
                )

            elif all(info.has_type(Json) for info in informations):
                # If the information is of type json, we use the json embedding function:
                json_strs = [
                    (
                        json.dumps(info.content)
                        if isinstance(info.content, dict)
                        else info.content
                    )
                    for info in informations
                ]

                # Embed the json:
                embeddings = self.api.embed_texts(
                    texts=json_strs, dimensions=self.embedding_length
                )

            elif all(info.has_type(Object) for info in informations):
                # If the information is of type object, we use the object embedding function:

                # First let's check that the object is a Pydantic BaseModel:
                if any(
                    not isinstance(info.content, BaseModel) for info in informations
                ):
                    raise ValueError(
                        "Information content must be a Pydantic BaseModel for object type Object."
                    )

                # Convert to JSON strings:
                json_strs = [
                    json.dumps(info.content.model_dump(mode="json"))
                    for info in informations
                ]

                # Embed the objects:
                embeddings = self.api.embed_texts(
                    texts=json_strs, dimensions=self.embedding_length
                )

            elif all(info.has_type(Image) for info in informations):
                # If the information is of type image, we use the image embedding function:
                image_uris = [info.content for info in informations]

                # Embed the images:
                embeddings = self.api.embed_images(
                    image_uris=image_uris, dimensions=self.embedding_length
                )
            elif all(info.has_type(Audio) for info in informations):
                # If the information is of type audio, we use the audio embedding function:
                audio_uris = [info.content for info in informations]

                # Embed the audio files:
                embeddings = self.api.embed_audios(
                    audio_uris=audio_uris, dimensions=self.embedding_length
                )
            elif all(info.has_type(Video) for info in informations):
                # If the information is of type video, we use the video embedding function:
                video_uris = [info.content for info in informations]

                # Embed the videos:
                embeddings = self.api.embed_videos(
                    video_uris=video_uris, dimensions=self.embedding_length
                )

            elif all(info.has_type(Document) for info in informations):
                # If the information is of type document, we use the document embedding function:
                document_uris = [info.content for info in informations]

                # Embed the videos:
                embeddings = self.api.embed_documents(
                    document_uris=document_uris, dimensions=self.embedding_length
                )

            elif all(info.has_type(Table) for info in informations):

                # First let's check that the all informations have a URIs:
                if any(not is_uri(info.content) for info in informations):
                    raise ValueError(
                        "All informations content must be a URI for table type."
                    )

                # Convert to Markdown:
                table_markdowns = [info.media.to_markdown() for info in informations]

                # Embed the table:
                embeddings = self.api.embed_texts(
                    texts=table_markdowns, dimensions=self.embedding_length
                )

            else:
                # If the information is of an unknown type, we raise an error:
                raise ValueError(f"Unknown information type: {informations[0].type}.")

        return embeddings

    def __del__(self):
        """Ensure the connection is closed when the object is garbage collected."""
        try:
            self.close()
        except:
            pass



from abc import ABC
from typing import Any, List, Sequence

from litemind.agent.messages.message import Message


class BaseApiCallbacks(ABC):
    """Abstract base class defining callback hooks for API operations.

    Subclass this and override methods to receive notifications about
    API events such as text generation, embedding, media description,
    and model selection.
    """

    def on_availability_check(self, available: bool) -> None:
        """
        Called when the availability and credentials check is performed.

        Parameters
        ----------
        available : bool
            Whether the API is available.
        """
        pass

    def on_model_list(self, models: List[str], **kwargs) -> None:
        """
        Called when the list of models is retrieved.

        Parameters
        ----------
        models : List[str]
            The list of models.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_best_model_selected(self, model_name: str, **kwargs) -> None:
        """
        Called when the best model is selected.

        Parameters
        ----------
        model_name : str
            The model name.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_text_generation(
        self, messages: List[Message], response: Message, **kwargs
    ) -> None:
        """
        Called when a text completion is generated.

        Parameters
        ----------
        messages : List[Message]
            The list of messages.
        response : Message
            The response message.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_text_streaming(self, fragment: str, **kwargs) -> None:
        """
        Called when a streaming text fragment is received during generation.

        Parameters
        ----------
        fragment : str
            Streaming fragment during generation.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_audio_transcription(
        self, transcription: str, audio_uri: str, **kwargs
    ) -> None:
        """
        Called when an audio transcription is generated.

        Parameters
        ----------
        transcription : str
            The transcription.
        audio_uri : str
            The audio URI.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_document_conversion(
        self, document_uri: str, markdown: str, **kwargs
    ) -> None:
        """
        Called when a document is converted to markdown.

        Parameters
        ----------
        document_uri : str
            The document URI.
        markdown : str
            The markdown content.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_video_conversion(
        self, video_uri: str, images: List[str], audio: str, **kwargs
    ) -> None:
        """
        Called when a video is converted to images and audio.

        Parameters
        ----------
        video_uri : str
            The video URI.
        images : List[str]
            The list of image URIs.
        audio : str
            The audio URI.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_audio_generation(self, text: str, audio_uri: str, **kwargs) -> None:
        """
        Called when an audio file is generated.

        Parameters
        ----------
        text : str
            The text prompt.
        audio_uri : str
            The audio URI.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_image_generation(self, prompt: str, image: Any, **kwargs) -> None:
        """
        Called when an image is generated.

        Parameters
        ----------
        prompt : str
            The image generation prompt.
        image : Any
            The generated image.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_text_embedding(
        self, texts: List[str], embeddings: Sequence[Sequence[float]], **kwargs
    ) -> None:
        """
        Called when text embeddings are generated.

        Parameters
        ----------
        texts : List[str]
            The input texts.
        embeddings : Sequence[Sequence[float]]
            The generated embeddings.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_image_embedding(
        self, image_uris: List[str], embeddings: Sequence[Sequence[float]], **kwargs
    ) -> None:
        """
        Called when image embeddings are generated.

        Parameters
        ----------
        image_uris : List[str]
            The input image URIs.
        embeddings : Sequence[Sequence[float]]
            The generated embeddings.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_audio_embedding(
        self, audio_uris: List[str], embeddings: Sequence[Sequence[float]], **kwargs
    ) -> None:
        """
        Called when audio embeddings are generated.

        Parameters
        ----------
        audio_uris : List[str]
            The input audio URIs.
        embeddings : Sequence[Sequence[float]]
            The generated embeddings.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_video_embedding(
        self, video_uris: List[str], embeddings: Sequence[Sequence[float]], **kwargs
    ) -> None:
        """
        Called when video embeddings are generated.

        Parameters
        ----------
        video_uris : List[str]
            The input video URIs.
        embeddings : Sequence[Sequence[float]]
            The generated embeddings.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_document_embedding(
        self, document_uris: List[str], embeddings: Sequence[Sequence[float]], **kwargs
    ) -> None:
        """
        Called when document embeddings are generated.

        Parameters
        ----------
        document_uris : List[str]
            The input document URIs.
        embeddings : Sequence[Sequence[float]]
            The generated embeddings.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_image_description(self, image_uri: str, description: str, **kwargs) -> None:
        """
        Called when an image is described.

        Parameters
        ----------
        image_uri : str
            The image URI.
        description : str
            The description.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_audio_description(self, audio_uri: str, description: str, **kwargs) -> None:
        """
        Called when an audio is described.

        Parameters
        ----------
        audio_uri : str
            The audio URI.
        description : str
            The description.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_video_description(self, video_uri: str, description: str, **kwargs) -> None:
        """
        Called when a video is described.

        Parameters
        ----------
        video_uri : str
            The video URI.
        description : str
            The description.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass

    def on_document_description(
        self, document_uri: str, description: str, **kwargs
    ) -> None:
        """
        Called when a document is described.

        Parameters
        ----------
        document_uri : str
            The document URI.
        description : str
            The description.
        kwargs : dict
            Other and additional keyword arguments.
        """
        pass



from typing import List, Tuple, Type

from litemind.apis.base_api import BaseApi
from litemind.apis.model_features import ModelFeatures
from litemind.media.conversion.converters.base_converter import BaseConverter
from litemind.media.media_base import MediaBase
from litemind.media.types.media_audio import Audio
from litemind.media.types.media_document import Document
from litemind.media.types.media_image import Image
from litemind.media.types.media_text import Text
from litemind.media.types.media_video import Video


class MediaConverterApi(BaseConverter):
    """Converts Image, Audio, Video, and Document media to Text via a LLM provider API.

    Delegates to the provider's ``describe_*`` methods (e.g.
    ``describe_image``, ``describe_audio``) to generate text descriptions.
    Only supports media types for which the underlying model has the
    corresponding feature (Image, Audio, Video, Document).
    """

    def __init__(self, api: BaseApi):
        """Initialise the converter with a litemind provider API.

        Parameters
        ----------
        api : BaseApi
            The provider API used to generate text descriptions of media.
        """

        super().__init__()
        self.api: BaseApi = api

    def rule(self) -> List[Tuple[Type[MediaBase], List[Type[MediaBase]]]]:
        """Declare all theoretically possible API-based media conversions.

        Returns a static list of all media-to-text conversions that the
        API *could* perform. The actual feature check is deferred to
        ``can_convert()`` which is called at conversion time, not during
        graph construction. This avoids a circular dependency where
        ``rule()`` calls ``has_model_support_for()`` which calls
        ``can_convert_within()`` which calls ``rule()`` again.

        Returns
        -------
        List[Tuple[Type[MediaBase], List[Type[MediaBase]]]]
            Rules for each media type the API could potentially describe.
        """
        return [
            (Image, [Text]),
            (Audio, [Text]),
            (Video, [Text]),
            (Document, [Text]),
        ]

    def can_convert(self, media: MediaBase) -> bool:
        """Check whether the API can describe the given media type.

        Parameters
        ----------
        media : MediaBase
            The media to check.

        Returns
        -------
        bool
            True if the media type is supported by the underlying model.
        """
        if isinstance(media, Image) and self.api.has_model_support_for(
            ModelFeatures.Image
        ):
            return True
        elif isinstance(media, Audio) and self.api.has_model_support_for(
            ModelFeatures.Audio
        ):
            return True
        elif isinstance(media, Video) and self.api.has_model_support_for(
            ModelFeatures.Video
        ):
            return True
        elif isinstance(media, Document) and self.api.has_model_support_for(
            ModelFeatures.Document
        ):
            return True
        else:
            return False

    def convert(self, media: MediaBase) -> List[MediaBase]:
        """Generate a text description of the media using the LLM API.

        Falls back to returning the original media if the API call fails.

        Parameters
        ----------
        media : MediaBase
            The media to describe.

        Returns
        -------
        List[MediaBase]
            A single-element list containing a Text description, or the
            original media if an error occurs.
        """

        try:
            if isinstance(media, Image) and self.api.has_model_support_for(
                ModelFeatures.Image
            ):
                description = self.api.describe_image(media.uri)
            elif isinstance(media, Audio) and self.api.has_model_support_for(
                ModelFeatures.Audio
            ):
                description = self.api.describe_audio(media.uri)
            elif isinstance(media, Video) and self.api.has_model_support_for(
                ModelFeatures.Video
            ):
                description = self.api.describe_video(media.uri)
            elif isinstance(media, Document) and self.api.has_model_support_for(
                ModelFeatures.Document
            ):
                description = self.api.describe_document(media.uri)
            else:
                raise ValueError(
                    f"Expected Image, Audio, Video, or Document media, got {type(media)}"
                )

            # Create a Text media with the description
            text_media = Text(description)

            return [text_media]

        except Exception:

            # Log the error
            import traceback

            traceback.print_exc()

            # If any error occurs, return the original media
            return [media]



import copy
import fnmatch
import os
from string import Formatter
from types import MappingProxyType
from typing import Any, List, Mapping, Optional, Sequence, Set, Type, Union

from arbol import aprint
from pydantic import BaseModel

from litemind.agent.messages.actions.tool_call import ToolCall
from litemind.agent.messages.actions.tool_use import ToolUse
from litemind.agent.messages.message_block import MessageBlock
from litemind.media.media_base import MediaBase
from litemind.media.types.media_action import Action
from litemind.media.types.media_audio import Audio
from litemind.media.types.media_code import Code
from litemind.media.types.media_document import Document
from litemind.media.types.media_file import File
from litemind.media.types.media_image import Image
from litemind.media.types.media_json import Json
from litemind.media.types.media_object import Object
from litemind.media.types.media_table import Table
from litemind.media.types.media_text import Text
from litemind.media.types.media_video import Video
from litemind.utils.extract_archive import extract_archive
from litemind.utils.file_types.file_extensions import OTHER_BINARY_EXTS
from litemind.utils.file_types.file_types import (
    has_extension,
    is_archive_file,
    is_audio_file,
    is_document_file,
    is_executable_file,
    is_image_file,
    is_prog_code_file,
    is_script_file,
    is_text_file,
    is_video_file,
    is_web_file,
)
from litemind.utils.folder_description import file_info_header, generate_tree_structure
from litemind.utils.normalise_uri_to_local_file_path import uri_to_local_file_path
from litemind.utils.uri_utils import is_uri


class Message:
    def __init__(
        self,
        text: Optional[str] = None,
        json: Optional[str] = None,
        obj: Optional[BaseModel] = None,
        image: Optional[str] = None,
        audio: Optional[str] = None,
        video: Optional[str] = None,
        document: Optional[str] = None,
        table: Optional[str] = None,
        role: str = "user",
    ):
        """
        Create a new message with optional initial content.

        Parameters
        ----------
        text : Optional[str]
            Text content to append as a text block.
        json : Optional[str]
            JSON string to append as a JSON block.
        obj : Optional[BaseModel]
            Pydantic model to append as an object block.
        image : Optional[str]
            Image URI to append as an image block.
        audio : Optional[str]
            Audio URI to append as an audio block.
        video : Optional[str]
            Video URI to append as a video block.
        document : Optional[str]
            Document URI to append as a document block.
        table : Optional[str]
            Table URI to append as a table block.
        role : str
            The role of the message sender. Must be one of 'system',
            'user', 'assistant', or 'tool'.

        Raises
        ------
        ValueError
            If ``role`` is not a valid role string.
        """

        # Check that the role is a string that is either: 'system', 'user', or 'assistant':
        if not isinstance(role, str):
            raise ValueError(f"Role must be a string, not {type(role)}")
        if role not in ["system", "user", "assistant", "tool"]:
            raise ValueError(
                f"Role must be 'system', 'user', or 'assistant', not {role}"
            )

        # Set the role of the message:
        self.role = role
        self.blocks: List[MessageBlock] = []

        # Append the text block if text is provided
        if text:
            self.append_text(text)
        if json:
            self.append_json(json)
        if obj:
            self.append_object(obj)
        if image:
            self.append_image(image)
        if audio:
            self.append_audio(audio)
        if video:
            self.append_video(video)
        if document:
            self.append_document(document)
        if table:
            self.append_table(table)

    def copy(self):
        """
        Create a copy of the message.

        Returns
        -------
        Message
            A copy of the message.
        """
        new_message = Message(role=self.role)
        new_message.blocks = self.blocks.copy()
        return new_message

    def __deepcopy__(self, memo):
        """
        Create a deep copy of the message. The blocks are also deep copied.

        Returns
        -------
        Message
            A deep copy of the message.
        """

        new_message = Message(role=self.role)
        new_message.blocks = [copy.deepcopy(block, memo) for block in self.blocks]
        return new_message

    def append_block(self, block: MessageBlock) -> MessageBlock:
        """
        Append a message block to the message.

        Parameters
        ----------
        block : MessageBlock
            The message block to append.

        Returns
        -------
        MessageBlock
            The appended message block.
        """
        self.blocks.append(block)
        return block

    def append_blocks(self, blocks: Union["Message", Sequence[MessageBlock]]):
        """
        Append multiple message blocks to the message.

        Parameters
        ----------
        blocks : Union[Message, Sequence[MessageBlock]]
            The message blocks to append. If a Message is provided, its
            blocks are extracted and appended.
        """

        # If the blocks are from another message, extract the blocks:
        if isinstance(blocks, Message):
            blocks = blocks.blocks

        # Append the blocks to the message:
        self.blocks.extend(blocks)

    def insert_block(
        self, block: MessageBlock, block_before: MessageBlock
    ) -> MessageBlock:
        """
        Insert a message block after an existing block.

        Parameters
        ----------
        block : MessageBlock
            The message block to insert.
        block_before : MessageBlock
            The existing block after which to insert the new block.

        Returns
        -------
        MessageBlock
            The inserted message block.
        """

        # find the index of the block_before:
        index = self.blocks.index(block_before)

        # insert the block after the block_before:
        self.blocks.insert(index + 1, block)

        return block

    def insert_blocks(
        self, blocks: List[MessageBlock], block_before: MessageBlock
    ) -> List[MessageBlock]:
        """
        Insert multiple message blocks after an existing block.

        Parameters
        ----------
        blocks : List[MessageBlock]
            The message blocks to insert.
        block_before : MessageBlock
            The existing block after which to insert.

        Returns
        -------
        List[MessageBlock]
            The inserted message blocks.
        """

        # insert the blocks after the block_before:
        for block in blocks:
            self.insert_block(block, block_before)
            block_before = block

        return blocks

    def insert_message(
        self, message: "Message", block_before: MessageBlock
    ) -> "Message":
        """
        Insert all blocks from another message after an existing block.

        Parameters
        ----------
        message : Message
            The message whose blocks will be inserted.
        block_before : MessageBlock
            The existing block after which to insert.

        Returns
        -------
        Message
            The inserted message.
        """

        self.insert_blocks(message.blocks, block_before)

        return message

    def append_media(self, media: MediaBase, **kwargs) -> MessageBlock:
        """
        Append a media object to the message as a new block.

        Parameters
        ----------
        media : MediaBase
            The media object to append.
        **kwargs
            Additional attributes for the message block.

        Returns
        -------
        MessageBlock
            The appended message block.
        """

        # Append the media block:
        return self.append_block(MessageBlock(media=media, **kwargs))

    def append_text(self, text: str, **kwargs) -> MessageBlock:
        """
        Append text to the message.

        Parameters
        ----------
        text : str
            The text to append.
        **kwargs
            Additional attributes for the text block.

        Returns
        -------
        MessageBlock
            The appended text block.
        """

        # Append the text block:
        return self.append_block(MessageBlock(media=Text(text=text), **kwargs))

    def append_quoted_text(
        self, text: str, quote_level: int = 1, **kwargs
    ) -> MessageBlock:
        """
        Append quoted text to the message using markdown blockquote syntax.

        Parameters
        ----------
        text : str
            The text to quote and append.
        quote_level : int
            The nesting level of the blockquote. Each level prepends an
            additional '>' character to each line.
        **kwargs
            Additional attributes for the text block.

        Returns
        -------
        MessageBlock
            The appended quoted text block.

        Raises
        ------
        ValueError
            If ``text`` is not a string.
        """

        # Check that it is really a string:
        if not isinstance(text, str):
            raise ValueError(f"Text must be a string, not {type(text)}")

        # Quote the text:
        quoted_text = "\n".join(
            f"{'>' * quote_level} {line}" for line in text.splitlines()
        )

        # Append the quoted text block:
        return self.append_block(MessageBlock(media=Text(text=quoted_text), **kwargs))

    def append_templated_text(self, template: str, **replacements) -> MessageBlock:
        """
        Append text from a template with placeholder substitution.

        Replaces all ``{placeholders}`` in *template* with values supplied
        via keyword arguments and appends the result as a new text block.

        Parameters
        ----------
        template : str
            The template string containing ``{name}`` placeholders.
        **replacements
            Keyword arguments whose names match the placeholders.

        Returns
        -------
        MessageBlock
            The appended text block with the filled template.

        Raises
        ------
        KeyError
            If one or more placeholders are missing a replacement.
        """
        # Discover every placeholder that appears in *template*
        required: set[str] = {
            field_name
            for _, field_name, _, _ in Formatter().parse(template)
            if field_name  # skip literal braces
        }

        missing = required.difference(replacements)
        if missing:
            raise KeyError(f"Missing replacement(s) for: {', '.join(sorted(missing))}")

        # MappingProxyType gives us an *immutable* view, guarding against
        # accidental mutation inside `format_map`.
        safe_mapping: Mapping[str, Any] = MappingProxyType(replacements)

        # Use `format_map` so that **replacements is not re-copied.
        filled_template = template.format_map(safe_mapping)

        # Append the filled template as a text block:
        return self.append_text(filled_template)

    def append_thinking(self, thinking_text: str, **kwargs) -> MessageBlock:
        """
        Append a thinking block to the message.

        Parameters
        ----------
        thinking_text: str
            The text to append to the message.
        kwargs: dict
            Additional attributes for the thinking block.

        Returns
        -------
        MessageBlock
            The thinking block appended to the message.

        """

        # Check that it is really a string:
        if not isinstance(thinking_text, str):
            raise ValueError(f"Text must be a string, not {type(thinking_text)}")

        # Add 'thinking' to the attributes:
        kwargs["thinking"] = True

        # Append the thinking block:
        return self.append_block(MessageBlock(media=Text(text=thinking_text), **kwargs))

    def append_json(
        self, json_str: str, source: Optional[str] = None, **kwargs
    ) -> MessageBlock:
        """
        Append a JSON string to the message.

        Parameters
        ----------
        json_str : str
            The JSON string to append.
        source : Optional[str]
            The source of the JSON (e.g., file path or URL).
        **kwargs
            Additional attributes for the JSON block.

        Returns
        -------
        MessageBlock
            The appended JSON block.
        """

        # Append the json block:
        return self.append_block(
            MessageBlock(media=Json(json_str), source=source, **kwargs)
        )

    def append_code(
        self, code: str, lang: str = "python", source: Optional[str] = None, **kwargs
    ) -> MessageBlock:
        """
        Append a code snippet to the message.

        Parameters
        ----------
        code : str
            The code to append.
        lang : str
            The programming language of the code.
        source : Optional[str]
            The source of the code (e.g., file path or URL).
        **kwargs
            Additional attributes for the code block.

        Returns
        -------
        MessageBlock
            The appended code block.
        """

        # Append the code block:
        return self.append_block(
            MessageBlock(media=Code(code=code, lang=lang), source=source, **kwargs)
        )

    def append_object(
        self, obj: BaseModel, source: Optional[str] = None, **kwargs
    ) -> MessageBlock:
        """
        Append a Pydantic model object to the message.

        Parameters
        ----------
        obj : BaseModel
            The Pydantic model instance to append.
        source : Optional[str]
            The source of the object (e.g., file path or URL).
        **kwargs
            Additional attributes for the object block.

        Returns
        -------
        MessageBlock
            The appended object block.
        """

        # Append the object block:
        return self.append_block(MessageBlock(media=Object(obj), **kwargs))

    def append_image(
        self, image_uri: str, source: Optional[str] = None, **kwargs
    ) -> MessageBlock:
        """
        Append an image to the message.

        Parameters
        ----------
        image_uri : str
            URI of the image (local file path or URL).
        source : Optional[str]
            The original source of the image.
        **kwargs
            Additional attributes for the image block.

        Returns
        -------
        MessageBlock
            The appended image block.
        """

        # Append the image block:
        return self.append_block(
            MessageBlock(media=Image(uri=image_uri), source=source, **kwargs)
        )

    def append_audio(
        self, audio_uri: str, source: Optional[str] = None, **kwargs
    ) -> MessageBlock:
        """
        Append an audio clip to the message.

        Parameters
        ----------
        audio_uri : str
            URI of the audio file (local file path or URL).
        source : Optional[str]
            The original source of the audio.
        **kwargs
            Additional attributes for the audio block.

        Returns
        -------
        MessageBlock
            The appended audio block.
        """

        # Append the audio block:
        return self.append_block(
            MessageBlock(media=Audio(uri=audio_uri), source=source, **kwargs)
        )

    def append_video(
        self, video_uri: str, source: Optional[str] = None, **kwargs
    ) -> MessageBlock:
        """
        Append a video to the message.

        Parameters
        ----------
        video_uri : str
            URI of the video file (local file path or URL).
        source : Optional[str]
            The original source of the video.
        **kwargs
            Additional attributes for the video block.

        Returns
        -------
        MessageBlock
            The appended video block.
        """

        # Append the video block:
        return self.append_block(
            MessageBlock(Video(uri=video_uri), source=source, **kwargs)
        )

    def append_document(
        self, document_uri: str, source: Optional[str] = None
    ) -> MessageBlock:
        """
        Append a document to the message.

        Parameters
        ----------
        document_uri : str
            URI of the document file (local file path or URL).
        source : Optional[str]
            The original source of the document.

        Returns
        -------
        MessageBlock
            The appended document block.
        """

        # Append the document block:
        return self.append_block(
            MessageBlock(Document(uri=document_uri), source=source)
        )

    def append_table(
        self, table: Union[str, "ndarray", "DataFrame"], source: Optional[str] = None
    ) -> MessageBlock:
        """
        Append a table to the message.

        Parameters
        ----------
        table : Union[str, ndarray, DataFrame]
            The table to append. Can be a URI of a local or remote file,
            a pandas DataFrame, a numpy array, or a Python list.
        source : Optional[str]
            The original source of the table (e.g., file path or URL).

        Returns
        -------
        MessageBlock
            The appended table block.

        Raises
        ------
        ValueError
            If ``table`` is not a supported type.
        """

        # Check that it is a numpy array or a pandas DataFrame:
        from numpy import ndarray
        from pandas import DataFrame

        # If table is an URI of table file (csv, tsv, xls, xlsx, etc.) download file and load it as a pandas DataFrame:
        if isinstance(table, str) and is_uri(table):

            # source is URI that the table comes from:
            source = table

            # Create a table block from the URI:
            block = MessageBlock(Table(table), source=source)

        elif isinstance(table, ndarray) or isinstance(table, list):

            # Convert numpy array to DataFrame:
            table = DataFrame(table)

            # Create a table block from the DataFrame:
            block = MessageBlock(Table.from_table(table), source=source)

        elif isinstance(table, DataFrame):

            # Create a table block from the DataFrame:
            block = MessageBlock(Table.from_table(table), source=source)

        else:
            raise ValueError(
                f"Table must be a numpy array or a pandas DataFrame, not {type(table)}"
            )

        # Append the table block:
        return self.append_block(block)

    def append_table_as_text(
        self, table: Union["ndarray", "DataFrame"], source: Optional[str] = None
    ):
        """
        Append a table to the message as a markdown text block.

        The table is converted to a markdown-formatted string before
        being appended as a text block.

        Parameters
        ----------
        table : Union[ndarray, DataFrame]
            The table to append. Can be a pandas DataFrame, a numpy
            array, list, or tuple.
        source : Optional[str]
            The original source of the table (unused, kept for API
            consistency).

        Returns
        -------
        MessageBlock
            The appended text block containing the table in markdown format.
        """

        # Convert the table to a markdown string and append it as text:
        markdown_string = Table.from_table(table).to_markdown()
        return self.append_text(markdown_string)

    def append_file(self, file_uri: str, source: Optional[str] = None) -> MessageBlock:
        """
        Append a raw file to the message.

        Parameters
        ----------
        file_uri : str
            The file URI to append. Must be a local ``file://`` URI.
        source : Optional[str]
            The original source of the file.

        Returns
        -------
        MessageBlock
            The appended file block.

        Raises
        ------
        ValueError
            If the URI does not start with ``file://`` or the file does
            not exist.
        """

        # Check that the file is a local file:
        if not file_uri.startswith("file://"):
            # Raise an exception if the file URI is not a local file:
            raise ValueError(
                f"File URI must be a local file URI that starts with 'file://', not '{file_uri}'"
            )

        # Remove the 'file://' prefix to get the local file path:
        file_uri_no_prefix = file_uri[7:]  # Remove 'file://' prefix

        # Check that the file URI is valid:
        if not os.path.exists(file_uri_no_prefix):
            raise ValueError(f"File '{file_uri_no_prefix}' does not exist.")

        # Append the file block:
        return self.append_block(MessageBlock(media=File(uri=file_uri), source=source))

    def append_folder(
        self,
        folder: str,
        depth: Optional[int] = None,
        allowed_extensions: List[str] = None,
        excluded_files: List[str] = None,
        all_archive_files: bool = False,
        include_hidden_files: bool = False,
        append_tree_structure: bool = True,
        date_and_times: bool = False,
        file_sizes: bool = True,
    ):
        """
        Append a folder to the message.

        Parameters
        ----------
        folder : str
            The folder to append.
        depth : Optional[int]
            The depth to traverse the folder.
            If None then there is no depth limits (default: None).
        allowed_extensions : List[str]
            The list of allowed file extensions (default: None).
        excluded_files : List[str]
            The list of excluded files (default: None).
        all_archive_files: bool
            Whether to include all files in archives,
            disregarding the depth of files in the archives (default: False).
        include_hidden_files: bool
            Whether to include hidden files (starting with '.') (default: False).
        append_tree_structure: bool
            Whether to append the tree structure of the folder to the message (default: True).
        date_and_times: bool
            Whether to include date and time information in the file info header (default: False).
        file_sizes: bool
            Whether to include file sizes in the file info header (default: True).
        """

        # if depth is zero, then we do not traverse the folder:
        if depth == 0:
            self.append_text(
                f"Folder '{folder}' is not traversed because depth is set to 0."
            )
            return

        # Expand folder string into an absolute path:
        folder = os.path.abspath(folder)

        # If folder does not exist, then append a message that explains that:
        if not os.path.exists(folder):
            self.append_text(f"Folder '{folder}' does not exist.")
            return

        # if folder is empty (does not contain any files or folders) then append a message that explains that:
        if os.path.isdir(folder) and not os.listdir(folder):
            self.append_text(f"Folder '{folder}' is empty.")
            return

        # If folder is a file, then append the file:
        if os.path.isfile(folder):
            # Append the file to the message:
            self.append_text(f"File '{folder}' is not a folder.")
            return

        if append_tree_structure:
            # 1) Generate and append the directory tree (with sizes, no timestamps).
            tree_structure = generate_tree_structure(
                folder,
                allowed_extensions=allowed_extensions,
                excluded_files=excluded_files,
                include_hidden_files=include_hidden_files,
                depth=depth,
            )

            # Append the tree structure to the message:
            self.append_text(f"Directory structure:\n{tree_structure}")

        # 2) Get immediate contents of the folder
        try:
            folder_contents = os.listdir(folder)
            files = []
            dirs = []

            # Separate files and directories
            for item in folder_contents:
                item_path = os.path.join(folder, item)
                if os.path.isfile(item_path):
                    files.append(item)
                elif os.path.isdir(item_path):
                    dirs.append(item)

            # Process files
            for file_name in sorted(files):
                # Skip hidden files if include_hidden_files is False
                if not include_hidden_files and (
                    file_name.startswith(".") or file_name.startswith("__")
                ):
                    continue

                # Skip files that are excluded:
                if excluded_files and any(
                    fnmatch.fnmatch(file_name, p) for p in excluded_files
                ):
                    aprint(f"File '{file_name}' is excluded.")
                    continue

                # Only keep files with allowed extensions:
                if allowed_extensions and not any(
                    file_name.endswith(ext) for ext in allowed_extensions
                ):
                    aprint(
                        f"File '{file_name}' is not allowed by the extensions filter."
                    )
                    continue

                # Get the file path and URI
                file_path = os.path.join(folder, file_name)
                file_uri = "file://" + file_path

                # if file is empty then just append 'Empty file' message:
                if os.stat(file_path).st_size == 0:
                    header = file_info_header(
                        file_path,
                        "Empty",
                        date_and_times=date_and_times,
                        file_sizes=file_sizes,
                    )
                    self.append_text(header + "\n")
                elif is_prog_code_file(file_path):
                    self.append_text(
                        file_info_header(
                            file_path,
                            "Prog. Lang. Code",
                            date_and_times=date_and_times,
                            file_sizes=file_sizes,
                        )
                    )
                    self.append_document(file_uri, source=file_path)
                elif is_script_file(file_path):
                    self.append_text(
                        file_info_header(
                            file_path,
                            "Script",
                            date_and_times=date_and_times,
                            file_sizes=file_sizes,
                        )
                    )
                    self.append_document(file_uri, source=file_path)
                elif is_web_file(file_path):
                    self.append_text(
                        file_info_header(
                            file_path,
                            "Web",
                            date_and_times=date_and_times,
                            file_sizes=file_sizes,
                        )
                    )
                    self.append_document(file_uri, source=file_path)
                elif is_text_file(file_path):
                    self.append_text(
                        file_info_header(
                            file_path,
                            "Text",
                            date_and_times=date_and_times,
                            file_sizes=file_sizes,
                        )
                    )
                    self.append_document(file_uri, source=file_path)
                elif is_image_file(file_path):
                    self.append_text(
                        file_info_header(
                            file_path,
                            "Image",
                            date_and_times=date_and_times,
                            file_sizes=file_sizes,
                        )
                    )
                    self.append_image(file_uri, source=file_path)
                elif is_audio_file(file_path):
                    self.append_text(
                        file_info_header(
                            file_path,
                            "Audio",
                            date_and_times=date_and_times,
                            file_sizes=file_sizes,
                        )
                    )
                    self.append_audio(file_uri, source=file_path)
                elif is_video_file(file_path):
                    self.append_text(
                        file_info_header(
                            file_path,
                            "Video",
                            date_and_times=date_and_times,
                            file_sizes=file_sizes,
                        )
                    )
                    self.append_video(file_uri, source=file_path)
                elif is_document_file(file_path):
                    self.append_text(
                        file_info_header(
                            file_path,
                            "Document",
                            date_and_times=date_and_times,
                            file_sizes=file_sizes,
                        )
                    )
                    self.append_document(file_uri, source=file_path)
                elif is_archive_file(file_path):
                    self.append_text(
                        file_info_header(
                            file_path,
                            "Archive (depth={depth})",
                            date_and_times=date_and_times,
                            file_sizes=file_sizes,
                        )
                    )
                    next_depth = None if all_archive_files or not depth else depth - 1
                    self.append_archive(file_uri, next_depth)
                elif is_executable_file(file_path):
                    self.append_text(
                        file_info_header(
                            file_path,
                            "Executable",
                            date_and_times=date_and_times,
                            file_sizes=file_sizes,
                        )
                    )
                    self.append_file(file_uri, source=file_path)
                elif has_extension(file_path, OTHER_BINARY_EXTS):
                    self.append_text(
                        file_info_header(
                            file_path,
                            "Binary",
                            date_and_times=date_and_times,
                            file_sizes=file_sizes,
                        )
                    )
                    self.append_file(file_path, source=file_path)
                else:
                    self.append_text(
                        file_info_header(
                            file_path,
                            "Other",
                            date_and_times=date_and_times,
                            file_sizes=file_sizes,
                        )
                    )
                    self.append_file(file_path, source=file_path)

            # Process subdirectories if depth allows
            if depth is None or depth > 0:
                next_depth = None if depth is None else depth - 1
                for folder_name in sorted(dirs):
                    # Skip hidden directories if include_hidden_files is False
                    if not include_hidden_files and (
                        folder_name.startswith(".") or folder_name.startswith("__")
                    ):
                        continue

                    # Skip directories that are excluded
                    if excluded_files and any(
                        fnmatch.fnmatch(folder_name, p) for p in excluded_files
                    ):
                        aprint(f"Folder '{folder_name}' is excluded.")
                        continue

                    self.append_text(f"\n###### Sub-folder: {folder_name}\n")
                    self.append_folder(
                        folder=os.path.join(folder, folder_name),
                        depth=next_depth,
                        allowed_extensions=allowed_extensions,
                        excluded_files=excluded_files,
                        all_archive_files=all_archive_files,
                        include_hidden_files=include_hidden_files,
                        append_tree_structure=False,
                    )
        except Exception as e:
            # print stack trace for debugging:
            import traceback

            traceback.print_exc()
            # If there is an error accessing the folder, append an error message:
            self.append_text(f"Error accessing folder '{folder}': {str(e)}")

    def append_archive(self, archive_uri: str, depth: int = -1):
        """
        Append an archive to the message.

        Parameters
        ----------
        archive_uri: str
            The archive's uri to append.
        depth: int
            The depth to traverse the archive (default: -1).


        """

        try:
            # Normalize the archive URI:
            archive_local_path = uri_to_local_file_path(archive_uri)

            # extract archive to temporary folder:
            temp_folder = extract_archive(archive_local_path)

            # append the extracted folder to the message:
            self.append_text(
                f"\n##### Contents of archive: {archive_uri} decompressed into folder: "
                + temp_folder
                + "\n"
            )

            # append the extracted folder to the message:
            self.append_folder(temp_folder, depth)
        except Exception as e:
            # If there is an error extracting the archive, append an error message:
            self.append_text(
                f"Error while attempting to extract files from archive '{archive_uri}': {str(e)}"
            )
            self.append_file(archive_uri, source=archive_uri)
            return

    def append_tool_call(
        self, tool_name: str, arguments: dict, id: str
    ) -> MessageBlock:
        """
        Append a tool call to the message.

        This is typically part of an assistant's response requesting that
        a tool be executed.

        Parameters
        ----------
        tool_name : str
            The name of the tool to call.
        arguments : dict
            The arguments for the tool call.
        id : str
            A unique identifier for this tool call.

        Returns
        -------
        MessageBlock
            The appended tool call block.
        """

        # Create a new tool use message block:
        block = MessageBlock(media=Action(ToolCall(tool_name, arguments, id)))

        # Append the tool use block to the message:
        return self.append_block(block)

    def append_tool_use(
        self, tool_name: str, arguments: dict, result: Any, id: str
    ) -> MessageBlock:
        """
        Append a tool use result to the message.

        This is typically part of a user message sent in reply to an
        assistant's tool call, containing the execution result.

        Parameters
        ----------
        tool_name : str
            The name of the tool that was executed.
        arguments : dict
            The arguments that were passed to the tool.
        result : Any
            The result returned by the tool.
        id : str
            The identifier matching the original tool call.

        Returns
        -------
        MessageBlock
            The appended tool use block.
        """

        # Create a new tool use message block:
        block = MessageBlock(media=Action(ToolUse(tool_name, arguments, result, id)))

        # Append the tool use block to the message:
        return self.append_block(block)

    def extract_markdown_block(
        self, filters: Union[str, List[str]], remove_quotes=True
    ) -> List[MessageBlock]:
        """
        Extract fenced code blocks from text blocks matching filter strings.

        Searches text blocks for the filter string, then extracts any
        fenced code blocks (delimited by triple backticks) found within
        matching blocks.

        Parameters
        ----------
        filters : Union[str, List[str]]
            Filter string(s) to search for. A block must contain at least
            one filter string to be searched for code blocks.
        remove_quotes : bool
            If True, strips the opening fence line (e.g., ````` ```python`````)
            from the extracted content.

        Returns
        -------
        List[MessageBlock]
            The extracted code blocks as text message blocks.
        """
        # Extract markdown blocks from the message that contain the given filter string.

        # If the filters parameter is a string, convert it to a list:
        if isinstance(filters, str):
            filters = [filters]

        # Initialize the list of markdown blocks:
        markdown_blocks: List[MessageBlock] = []

        # Iterate over the blocks in the message:
        for block in self.blocks:

            # Check if the block is a text block and contains the filter string:
            if isinstance(block.media, Text) and any(
                f in block.get_content() for f in filters
            ):
                # Parse the string to extract text strings in the form: "```... ```"
                # and create a new markdown block for each one.
                text = "" + block.get_content()

                # Extract markdown blocks from the text:
                while "```" in text:
                    start = text.find("```")
                    end = text.find("```", start + 3)
                    if end == -1:
                        break
                    if remove_quotes:
                        # find the first '\n' after the first ``` and set start after '\n':
                        start = text.find("\n", start + 3) + 1
                        # add the markdown block to the list:
                        markdown_blocks.append(
                            MessageBlock(media=Text(text[start:end]))
                        )
                    else:
                        # add the markdown block to the list:
                        markdown_blocks.append(
                            MessageBlock(media=Text(text[start : end + 3]))
                        )

                    # remove the markdown block from the text:
                    text = text[end + 3 :]

        return markdown_blocks

    def list_media_types(self) -> Sequence[Type[MediaBase]]:
        """
        List the media classes present in the message blocks.

        Returns
        -------
        Sequence[Type[MediaBase]]
            A list of media classes present in the message blocks.
        """
        media_classes: Set[Type[MediaBase]] = set()
        for block in self.blocks:
            if isinstance(block.media, MediaBase):
                media_classes.add(type(block.media))
        return list(media_classes)

    @staticmethod
    def list_media_types_in(messages: Sequence["Message"]) -> Sequence[Type[MediaBase]]:
        """
        List the media classes present in the given messages.

        Parameters
        ----------
        messages : Sequence[Message]
            The messages to analyze.

        Returns
        -------
        Sequence[Type[MediaBase]]
            A list of media classes present in the messages.
        """
        media_classes: Set[Type[MediaBase]] = set()
        message: Message
        for message in messages:
            media_type: Type[MediaBase]
            for media_type in message.list_media_types():
                media_classes.add(media_type)
        return list(media_classes)

    def convert_media(
        self,
        allowed_media_types: Optional[List[Type[MediaBase]]] = None,
        media_converter: Optional["MediaConverter"] = None,
    ) -> "Message":
        """
        Convert the media in the message blocks to the allowed media types.

        Parameters
        ----------
        allowed_media_types : Optional[List[Type[MediaBase]]]
            The list of allowed media types to convert to. If None, only text media types are allowed.
        media_converter: Optional[MediaConverter]
            The media converter to use for conversion. If None, a default media converter is used.

        Returns
        -------
        Message
            A new message with the converted media blocks.
        """

        if allowed_media_types is None:
            # If no allowed media types are provided, default to text media type:
            allowed_media_types = [Text]

        # If no media converter is provided, instantiate a default one:
        if media_converter is None:
            from litemind.media.conversion.media_converter import MediaConverter

            # Instantiate the media converter:
            media_converter = MediaConverter()

            # Add the media converter supporting APIs. This converter can use any model from the API that supports the required features.
            media_converter.add_default_converters()

        # Convert the message using the media converter:
        converted_messages = media_converter.convert(
            [self], allowed_media_types=allowed_media_types
        )

        # Get the converted message:
        converted_message = converted_messages[0]

        return converted_message

    def compress_text(
        self, text_compressor: Optional["TextCompressor"] = None
    ) -> "Message":
        """
        Compress the text in the message blocks by using TextCompressor
        to reduce the size of the text blocks.

        Parameters
        ----------
        text_compressor: Optional[TextCompressor]
            The text compressor to use for compression. If None, a default text compressor is used.

        Returns
        -------
        Message
            A new message with the compressed text blocks.
        """

        # Create a new message to hold the compressed text blocks:
        compressed_message = Message(role=self.role)

        # Create a text compressor:
        if text_compressor is None:
            # If no text compressor is provided, instantiate a default one:
            from litemind.utils.text_compressor import TextCompressor

            text_compressor = TextCompressor()

        # Iterate over the blocks in the message:
        for block in self.blocks:
            if isinstance(block.media, Text):
                # Compress the text block:
                compressed_text = text_compressor.compress(block.media.text)

                if len(compressed_text) > 0:
                    # Append the compressed text block to the compressed message:
                    compressed_message.append_text(compressed_text, **block.attributes)
            else:
                # Append the block as is if it is not a text block:
                compressed_message.append_block(block)

        return compressed_message

    def __getitem__(self, index: int) -> Union[MessageBlock, None]:
        """
        Get the message block at the given index.

        Parameters
        ----------
        index : int
            The index of the message block to get.

        Returns
        -------
        MessageBlock or None
            The message block at the given index, or None if the index is out of range.
        """
        return self.blocks[index]

    def __len__(self) -> int:
        """
        Return the number of blocks in the message.

        Returns
        -------
        int
            The number of blocks in the message.
        """
        return len(self.blocks)

    def __contains__(self, item: str) -> bool:
        """
        Check if the given string is in the message blocks.

        Parameters
        ----------
        item : str
            The string to check for.

        Returns
        -------
        bool
            True if the string is found, False otherwise.
        """
        for block in self.blocks:
            if item in str(block):
                return True
        return False

    def has(self, block_type: Type[MediaBase]):
        """
        Check if the given block type is in the message blocks.

        Parameters
        ----------
        block_type : Type[MediaBase]
            The block type to check for.

        Returns
        -------
        bool
            True if the block type is found, False otherwise.
        """
        for block in self.blocks:
            if isinstance(block.media, block_type):
                return True
        return False

    def __str__(self) -> str:
        """
        Return a human-readable string representation of the message.

        Returns
        -------
        str
            The message prefixed with its role, followed by all blocks.
        """
        message_string = f"*{self.role}*:\n"
        for block in self.blocks:
            message_string += str(block) + "\n"
        return message_string

    def lower(self) -> str:
        """
        Return the message string representation in lowercase.

        Returns
        -------
        str
            The lowercased string representation.
        """
        return self.__str__().lower()

    def to_plain_text(self) -> str:
        """
        Return the message as plain text.

        Returns
        -------
        str
            The message as plain text.

        """
        plain_text = ""
        for block in self.blocks:
            plain_text += str(block.media) + "\n"
        return plain_text

    def to_markdown(self, media_converter: Optional["MediaConverter"] = None):
        """
        Convert the message to a markdown string.

        All media blocks are converted to text via the media converter,
        then concatenated into a single markdown string.

        Parameters
        ----------
        media_converter : Optional[MediaConverter]
            The media converter to use. If None, a default converter is
            created with default converters.

        Returns
        -------
        str
            The message content as a markdown string.
        """

        if media_converter is None:
            # If no media converter is provided, instantiate a default one:
            from litemind.media.conversion.media_converter import MediaConverter

            # Instantiate the media converter:
            media_converter = MediaConverter()

            # Add the media converter supporting APIs. This converter can use any model from the API that supports the required features.
            media_converter.add_default_converters()

        # Convert the document to markdown:
        from litemind.media.types.media_text import Text

        markdown_text = media_converter.convert(
            messages=[self], allowed_media_types=[Text]
        )

        # Get the converted message:
        converted_message = markdown_text[0]

        # Get the markdown string from the converted message blocks:
        markdown_string = converted_message.to_plain_text()

        return markdown_string

    def report(self, as_string=True) -> Union[dict, str]:
        """
        Generate a report with statistics about the message blocks.

        Parameters
        ----------
        as_string : bool
            Whether to return the report as a formatted string (True) or a dictionary (False).

        Returns
        -------
        Union[dict, str]
            A report containing statistics about the message blocks.
        """
        # Initialize report data
        report_data = {
            "total_blocks": len(self.blocks),
            "blocks_by_type": {},
            "longest_blocks": {},
            "average_block_size": 0,
            "average_by_type": {},
            "total_characters": 0,
            "role": self.role,
        }

        # Collect data for each block
        block_lengths = []
        lengths_by_type = {}

        for block in self.blocks:
            # Get media type name
            media_type = type(block.media).__name__

            # Count blocks by type
            if media_type not in report_data["blocks_by_type"]:
                report_data["blocks_by_type"][media_type] = 0
                lengths_by_type[media_type] = []

            report_data["blocks_by_type"][media_type] += 1

            # Calculate block length (character count)
            block_str = str(block)
            block_length = len(block_str)
            block_lengths.append(block_length)
            lengths_by_type[media_type].append(block_length)

            report_data["total_characters"] += block_length

            # Check if this is the longest block of its type
            if (
                media_type not in report_data["longest_blocks"]
                or block_length > report_data["longest_blocks"][media_type]["length"]
            ):
                # Preview the content of the block:
                preview_block_str = (
                    block_str[:100] if len(block_str) > 100 else block_str
                )

                # Replace newlines with emoji characters that look line new lines for better preview:
                preview_block_str = preview_block_str.replace("\n", "↩️")

                report_data["longest_blocks"][media_type] = {
                    "length": block_length,
                    "content_preview": preview_block_str + "...",
                }

        # Calculate averages
        if report_data["total_blocks"] > 0:
            report_data["average_block_size"] = (
                report_data["total_characters"] / report_data["total_blocks"]
            )

        # Calculate average by type
        for media_type, lengths in lengths_by_type.items():
            if lengths:
                report_data["average_by_type"][media_type] = sum(lengths) / len(lengths)

        # Add additional statistics
        if block_lengths:
            report_data["min_block_size"] = min(block_lengths)
            report_data["max_block_size"] = max(block_lengths)
            report_data["median_block_size"] = sorted(block_lengths)[
                len(block_lengths) // 2
            ]

        if not as_string:
            return report_data

        # Format report as string
        report_str = f"Message Report (Role: {self.role})\n"
        report_str += f"Total blocks: {report_data['total_blocks']}\n"
        report_str += f"Total characters: {report_data['total_characters']}\n\n"

        report_str += "Blocks by type:\n"
        for media_type, count in sorted(report_data["blocks_by_type"].items()):
            report_str += f"  {media_type}: {count} blocks\n"

        report_str += f"\nAverage block size: {report_data['average_block_size']:.2f} characters\n"

        report_str += "\nAverage size by type:\n"
        for media_type, avg in sorted(report_data["average_by_type"].items()):
            report_str += f"  {media_type}: {avg:.2f} characters\n"

        report_str += "\nLongest blocks by type:\n"
        for media_type, info in sorted(report_data["longest_blocks"].items()):
            report_str += f"  {media_type}: {info['length']} characters\n"
            report_str += f"    Preview: {info['content_preview']}\n"

        if "min_block_size" in report_data:
            report_str += (
                f"\nMin block size: {report_data['min_block_size']} characters\n"
            )
            report_str += (
                f"Max block size: {report_data['max_block_size']} characters\n"
            )
            report_str += (
                f"Median block size: {report_data['median_block_size']} characters\n"
            )

        return report_str

    def __repr__(self) -> str:
        """
        Return the string representation of the message.

        Returns
        -------
        str
            Same as ``__str__``.
        """
        return str(self)

    def __iter__(self):
        """
        Iterate over the message blocks.
        """
        return iter(self.blocks)

    def __hash__(self):
        """
        Return the hash of the message.

        Returns
        -------
        int
            Hash based on the role and blocks.
        """
        return hash((self.role, tuple(self.blocks)))

    def __eq__(self, other: "Message") -> bool:
        """
        Compare two messages for equality based on role and blocks.

        Parameters
        ----------
        other : Message
            The message to compare with.

        Returns
        -------
        bool
            True if both messages have the same role and identical blocks.

        Raises
        ------
        TypeError
            If ``other`` is not a Message instance.
        """

        # Check if is a Message:
        if not isinstance(other, Message):
            raise TypeError(f"Cannot compare {type(other)} with {type(self)}")

        # Check if the roles are the same:
        if self.role != other.role:
            return False

        # Check that there is the same number of blocks:
        if len(self.blocks) != len(other.blocks):
            return False

        # Check that the blocks are the same:
        for block1, block2 in zip(self.blocks, other.blocks):
            if block1 != block2:
                return False

        # If all checks pass, the messages are equal:
        return True

    def __ne__(self, other: "Message") -> bool:
        """
        Compare two messages for inequality.

        Parameters
        ----------
        other : Message
            The message to compare with.

        Returns
        -------
        bool
            True if the messages differ.
        """
        return not self == other

    def __add__(self, other: Union["Message", MessageBlock]) -> "Message":
        """
        Concatenate two messages.

        Parameters
        ----------
        other : Union[Message, MessageBlock]
            The message or message block to concatenate.

        Returns
        -------
        Message
            The concatenated message.
        """

        # Create a new message as copy of the current message:
        new_message = self.copy()

        if isinstance(other, Message):
            # Concatenate the blocks of the other message to the new message:
            new_message.blocks += other.blocks
        elif isinstance(other, MessageBlock):
            # Append the message block to the new message:
            new_message.blocks.append(other)
        else:
            # Raise an error if the other object is not a Message or MessageBlock:
            raise ValueError(
                f"Cannot concatenate message with {type(other)}. Only Message or MessageBlock are allowed."
            )

        # Return the new message:
        return new_message

    def __iadd__(self, other: Union["Message", MessageBlock]) -> "Message":
        """
        Concatenate another message to the current message.

        Parameters
        ----------
        other : Union[Message, MessageBlock]
            The message or message block to concatenate.

        Returns
        -------
        Message
            The concatenated message.
        """
        if isinstance(other, Message):
            # Concatenate the blocks of the other message to the current message:
            self.blocks += other.blocks
        elif isinstance(other, MessageBlock):
            # Append the message block to the current message:
            self.blocks.append(other)
        else:
            # Raise an error if the other object is not a Message or MessageBlock:
            raise ValueError(
                f"Cannot concatenate message with {type(other)}. Only Message or MessageBlock are allowed."
            )

        # Return the current message:
        return self

    def __radd__(self, other: Union["Message", MessageBlock]) -> "Message":
        """
        Concatenate the current message to another message.

        Parameters
        ----------
        other : Union[Message, MessageBlock]
            The message or message block to concatenate.

        Returns
        -------
        Message
            The concatenated message.
        """

        if isinstance(other, Message):
            # Create a copy of the other message and append self's blocks:
            new_message = other.copy()
            new_message.blocks += self.blocks
        elif isinstance(other, MessageBlock):
            # Create a copy of self and prepend the block:
            new_message = self.copy()
            new_message.blocks.insert(0, other)
        else:
            # Raise an error if the other object is not a Message or MessageBlock:
            raise ValueError(
                f"Cannot concatenate message with {type(other)}. Only Message or MessageBlock are allowed."
            )
        return new_message



"""
Litemind: A unified API wrapper and agentic AI framework.

Litemind provides a consistent interface to multiple LLM providers (OpenAI,
Anthropic, Google Gemini, Ollama) along with an agentic framework for building
tool-augmented, multimodal AI applications.

This top-level package exposes the main user-facing classes and lazy
availability checks for API backends and vector databases.
"""

__version__ = "2026.2.10.1"

from arbol import aprint, asection

# Core exports for user-facing API:
from litemind.agent.agent import Agent
from litemind.agent.augmentations.information.information import Information

# API class imports:
from litemind.agent.augmentations.vector_db.in_memory_vector_db import (
    InMemoryVectorDatabase,
)
from litemind.agent.augmentations.vector_db.qdrant_vector_db import QdrantVectorDatabase
from litemind.agent.messages.conversation import Conversation
from litemind.agent.messages.message import Message
from litemind.agent.tools.toolset import ToolSet
from litemind.apis.combined_api import CombinedApi
from litemind.apis.default_api import DefaultApi
from litemind.apis.model_features import ModelFeatures
from litemind.apis.providers.anthropic.anthropic_api import AnthropicApi
from litemind.apis.providers.google.google_api import GeminiApi
from litemind.apis.providers.ollama.ollama_api import OllamaApi
from litemind.apis.providers.openai.openai_api import OpenAIApi

# All API classes (static list for lazy initialization):
_ALL_API_CLASSES = [
    DefaultApi,
    OpenAIApi,
    OllamaApi,
    AnthropicApi,
    GeminiApi,
    CombinedApi,
]

# All vector database classes (static list for lazy initialization):
_ALL_VECDB_CLASSES = [QdrantVectorDatabase, InMemoryVectorDatabase]

# Cached results (None means not yet computed):
_api_implementations = None
_vecdb_implementations = None


def get_available_apis():
    """
    Get a list of available API implementations.

    Lazily checks availability and credentials of each registered API
    provider on the first call, then caches the result for subsequent
    calls. ``CombinedApi`` is included only when at least one concrete
    provider (other than ``DefaultApi``) is available.

    Returns
    -------
    list[type[BaseApi]]
        API classes whose providers are reachable and have valid
        credentials.
    """
    global _api_implementations

    if _api_implementations is not None:
        return _api_implementations

    # Build new list (don't modify _ALL_API_CLASSES)
    available = []

    for api_class in _ALL_API_CLASSES:
        # Skip CombinedApi initially - we'll handle it at the end
        if api_class == CombinedApi:
            continue

        try:
            with asection(f"Checking availability of: {api_class.__name__}"):
                # Try to instantiate the API class:
                api_instance = api_class()

                # Check if the API is available and credentials are correct:
                if (
                    not api_instance.check_availability_and_credentials()
                    or len(api_instance.list_models()) == 0
                ):
                    aprint(f"API {api_class.__name__} is not available. Skipping.")
                else:
                    aprint(
                        f"API {api_class.__name__} is available and has {len(api_instance.list_models())} models."
                    )
                    available.append(api_class)

        except Exception as e:
            aprint(
                f"API {api_class.__name__} could not be instantiated: {e}. Skipping."
            )

    # Add CombinedApi only if there are other APIs available
    # (need more than just DefaultApi)
    non_default_apis = [a for a in available if a != DefaultApi]
    if len(non_default_apis) > 0:
        available.append(CombinedApi)
    else:
        aprint("No individual APIs available. Skipping CombinedApi.")

    _api_implementations = available
    return _api_implementations


def get_available_vecdbs():
    """
    Get a list of available vector database implementations.

    Lazily checks that the required dependencies for each vector database
    backend are installed on the first call, then caches the result.
    ``QdrantVectorDatabase`` is excluded when no API provider is available
    because it relies on external embeddings.

    Returns
    -------
    list[type[BaseVectorDB]]
        Vector database classes whose dependencies are satisfied.
    """
    global _vecdb_implementations

    if _vecdb_implementations is not None:
        return _vecdb_implementations

    # Build new list (don't modify _ALL_VECDB_CLASSES)
    available = []

    for vecdb_class in _ALL_VECDB_CLASSES:
        try:
            # Check specific dependencies based on the vector database class
            if vecdb_class == QdrantVectorDatabase:
                import qdrant_client  # noqa: F401

            # InMemoryVectorDatabase doesn't require external dependencies
            available.append(vecdb_class)

        except ImportError as e:
            aprint(
                f"Vector database {vecdb_class.__name__} dependencies are not installed: {e}. Skipping."
            )
        except Exception as e:
            aprint(
                f"Error checking vector database {vecdb_class.__name__}: {e}. Skipping."
            )

    # If no APIs are available, QdrantVectorDatabase cannot be used
    # (it needs embeddings). Check lazily only when needed.
    if QdrantVectorDatabase in available:
        apis = get_available_apis()
        if len(apis) == 0:
            aprint("No API implementations available. Removing QdrantVectorDatabase.")
            available = [v for v in available if v != QdrantVectorDatabase]

    _vecdb_implementations = available
    return _vecdb_implementations


# For backwards compatibility, provide API_IMPLEMENTATIONS and VECDB_IMPLEMENTATIONS
# as static lists of all API/VecDB classes. These lists no longer filter by availability
# at import time (to avoid network calls during import).
# To get only available implementations, use get_available_apis() or get_available_vecdbs().
API_IMPLEMENTATIONS = _ALL_API_CLASSES.copy()
VECDB_IMPLEMENTATIONS = _ALL_VECDB_CLASSES.copy()

# Initialize and silence the Abseil logging system:
import absl.logging

absl.logging.set_verbosity(absl.logging.FATAL)
absl.logging.use_absl_handler()

__all__ = [
    # Version
    "__version__",
    # API classes
    "DefaultApi",
    "OpenAIApi",
    "OllamaApi",
    "AnthropicApi",
    "GeminiApi",
    "CombinedApi",
    # Vector databases
    "InMemoryVectorDatabase",
    "QdrantVectorDatabase",
    # Core agent API
    "Agent",
    "Message",
    "Conversation",
    "ToolSet",
    "ModelFeatures",
    "Information",
    # Availability functions
    "get_available_apis",
    "get_available_vecdbs",
    # Legacy (static lists)
    "API_IMPLEMENTATIONS",
    "VECDB_IMPLEMENTATIONS",
]



from typing import List, Optional, Sequence, Set, Type

from arbol import aprint

from litemind.agent.messages.message import Message
from litemind.agent.messages.message_block import MessageBlock
from litemind.media.conversion.converters.base_converter import BaseConverter
from litemind.media.media_base import MediaBase
from litemind.media.types.media_action import Action
from litemind.media.types.media_types import all_media_types


class MediaConverter:
    """Pipeline for converting messages with arbitrary media to a restricted set of allowed types.

    Maintains an ordered list of ``BaseConverter`` instances. When
    converting, the first converter that can handle a given media takes
    precedence. Conversion is applied recursively by default so that
    intermediate types (e.g. Table -> Text) are fully resolved.
    """

    def __init__(self):
        # List of media converters
        self.media_converters: List[BaseConverter] = []

    def add_default_converters(
        self,
        convert_audio: bool = True,
        convert_videos: bool = True,
        convert_documents: bool = True,
    ) -> List[BaseConverter]:
        """Register the standard set of converters.

        Adds converters for NdImage, Table, Object, Code, Json, and
        optionally Video (requires ffmpeg), Document (requires pymupdf
        and/or docling), and Audio (requires local Whisper). Converters
        are only added when their dependencies are available.

        Parameters
        ----------
        convert_audio : bool, optional
            Include the Whisper-based audio converter. Default is True.
        convert_videos : bool, optional
            Include the ffmpeg-based video converter. Default is True.
        convert_documents : bool, optional
            Include document converters (pymupdf, docling, txt fallback).
            Default is True.

        Returns
        -------
        List[BaseConverter]
            The converters that were actually added.
        """

        # Create a list of default converters
        default_converters = []

        # Import the default converters
        from litemind.media.conversion.converters.code_converter import CodeConverter
        from litemind.media.conversion.converters.document_converter_docling import (
            is_docling_available,
        )
        from litemind.media.conversion.converters.document_converter_pymupdf import (
            is_pymupdf_available,
        )
        from litemind.media.conversion.converters.document_converter_txt import (
            DocumentConverterTxt,
        )
        from litemind.media.conversion.converters.file_converter import FileConverter
        from litemind.media.conversion.converters.json_converter import JsonConverter
        from litemind.media.conversion.converters.ndimage_converter import (
            NdImageConverter,
        )
        from litemind.media.conversion.converters.object_converter import (
            ObjectConverter,
        )
        from litemind.media.conversion.converters.table_converter import TableConverter

        # Import the is_* functions:
        from litemind.utils.ffmpeg_utils import is_ffmpeg_available
        from litemind.utils.whisper_transcribe_audio import is_local_whisper_available

        # Add default converters to the list
        default_converters.append(NdImageConverter())
        default_converters.append(TableConverter())
        default_converters.append(ObjectConverter())
        default_converters.append(CodeConverter())
        default_converters.append(JsonConverter())

        if convert_videos and is_ffmpeg_available():
            from litemind.media.conversion.converters.video_converter_ffmpeg import (
                VideoConverterFfmpeg,
            )

            # Add the video converter to the list of default converters
            # This converter is only available if ffmpeg is installed
            default_converters.append(VideoConverterFfmpeg())

        if convert_documents and is_pymupdf_available():
            from litemind.media.conversion.converters.document_converter_pymupdf import (
                DocumentConverterPymupdf,
            )

            # Add the document converter to the list of default converters
            # This converter is only available if pymupdf is installed
            default_converters.append(DocumentConverterPymupdf())

        if convert_documents and is_docling_available():
            from litemind.media.conversion.converters.document_converter_docling import (
                DocumentConverterDocling,
            )

            # Add the document converter to the list of default converters
            # This converter is only available if docling is installed
            default_converters.append(DocumentConverterDocling())

        if convert_audio and is_local_whisper_available():
            from litemind.media.conversion.converters.audio_converter_whisper_local import (
                AudioConverterWhisperLocal,
            )

            # Add the audio converter to the list of default converters
            # This converter is only available if local whisper is installed
            default_converters.append(AudioConverterWhisperLocal())

        # This converter is at the end of the list to have the lowest priority
        if convert_documents:
            default_converters.append(DocumentConverterTxt())

        # Add FileConverter to the list of default converters
        default_converters.append(FileConverter())

        # Add default converters to the list of media converters
        self.media_converters.extend(default_converters)

        return default_converters

    def add_media_converter(
        self, media_converter: BaseConverter, highest_priority: bool = False
    ):
        """Add a converter to the pipeline.

        Parameters
        ----------
        media_converter : BaseConverter
            The converter to add.
        highest_priority : bool, optional
            If True, insert at the front of the list so it takes precedence
            over existing converters. Default is False (append).
        """
        if highest_priority:
            # Add the media converter to the beginning of the list
            self.media_converters.insert(0, media_converter)
        else:
            # Add the media converter to the end of the list
            self.media_converters.append(media_converter)

    def remove_media_converter(self, media_converter: BaseConverter):
        """Remove a converter from the pipeline.

        Parameters
        ----------
        media_converter : BaseConverter
            The converter instance to remove.

        Raises
        ------
        ValueError
            If the converter is not in the pipeline.
        """
        self.media_converters.remove(media_converter)

    def can_convert_within(
        self,
        source_media_type: Type[MediaBase],
        allowed_media_types: Set[Type[MediaBase]],
    ) -> bool:
        """Check whether a source type can be fully converted to allowed types.

        Performs a BFS over the conversion graph to verify that every
        reachable type from *source_media_type* is contained in
        *allowed_media_types*.

        Parameters
        ----------
        source_media_type : Type[MediaBase]
            The media type to start from.
        allowed_media_types : Set[Type[MediaBase]]
            The set of acceptable target types.

        Returns
        -------
        bool
            True if all conversion outputs are within the allowed set.
        """

        # If there are no converters, then we cannot convert anything, source type must be in allowed types:
        if not self.media_converters:
            return source_media_type in allowed_media_types

        # Build a conversion graph from all converter rules
        conversion_graph = {}

        # Process all converter rules to build graph
        for converter in self.media_converters:
            for src_type, target_types in converter.rule():
                if src_type not in conversion_graph:
                    conversion_graph[src_type] = set()
                # Add all possible conversion targets
                conversion_graph[src_type].update(target_types)

        # Queue for BFS traversal
        queue = [source_media_type]
        # Keep track of visited types to avoid cycles
        visited = set()

        # BFS to find all possible conversion results
        while queue:
            current_type = queue.pop(0)

            # Skip if already visited
            if current_type in visited:
                continue

            visited.add(current_type)

            # Check if current type can be converted
            if current_type in conversion_graph:
                for next_type in conversion_graph[current_type]:
                    # If any next type is not in allowed types, return False
                    if next_type not in allowed_media_types:
                        return False

                    if next_type not in visited:
                        queue.append(next_type)

        # If we get here, all conversions are within allowed types
        return True

    def get_convertible_media_types(
        self, allowed_media_types: Set[Type[MediaBase]]
    ) -> Set[Type[MediaBase]]:
        """Find all media types that can be converted to the allowed set.

        Parameters
        ----------
        allowed_media_types : Set[Type[MediaBase]]
            The set of acceptable target types.

        Returns
        -------
        Set[Type[MediaBase]]
            All media types for which ``can_convert_within`` returns True.
        """
        convertable_media_types = set()

        # iterate through all media types that derive from MediaBase:
        for media_type in all_media_types():

            # Check if the converter can convert this media type:
            if self.can_convert_within(media_type, allowed_media_types):
                # If it can, add it to the set of convertable media types
                convertable_media_types.add(media_type)

        return convertable_media_types

    def convert(
        self,
        messages: List[Message],
        allowed_media_types: Sequence[Type[MediaBase]],
        exclude_extensions: Optional[Sequence[str]] = None,
        recursive: bool = True,
    ) -> List[Message]:
        """Convert messages so they contain only allowed media types.

        Iterates over each message block and applies the first matching
        converter. The conversion is **not** performed in place; new
        messages are returned.

        Parameters
        ----------
        messages : List[Message]
            The messages to convert.
        allowed_media_types : Sequence[Type[MediaBase]]
            Media types that should be kept as-is.
        exclude_extensions : Sequence[str], optional
            File extensions to skip during conversion (media with these
            extensions are kept unchanged). Default is None (no exclusions).
        recursive : bool, optional
            If True, reapply conversion until no further changes occur.
            Default is True.

        Returns
        -------
        List[Message]
            New messages containing only allowed (or unconvertible) media.
        """

        # List that will hold the converted messages:
        converted_messages = []

        # By default no conversion happened:
        conversion_happened = False

        # Iterate over the messages:
        for message in messages:

            # Create a new message with the same role as the original message:
            new_message = Message(role=message.role)

            # Iterate over the blocks in the message:
            for block in message.blocks:

                # extract attributes from block:
                attributes = block.attributes

                # get the media from the block:
                media = block.media

                # Attempt to get the extension if defined:
                extension = getattr(media, "extension", None)

                if any(
                    isinstance(media, allowed_type)
                    for allowed_type in allowed_media_types
                ):
                    # If the media type is allowed, keep it as is:
                    new_message.append_block(block)
                elif (
                    extension is not None
                    and exclude_extensions is not None
                    and extension in exclude_extensions
                ):
                    # If the media type is not allowed and the extension is in the exclude list, skip conversion:
                    # aprint("Warning: Skipping media with excluded extension:", extension)
                    new_message.append_block(block)
                else:
                    # Otherwise, attempt to convert it using the available converters
                    converted = None

                    # Iterate over the media converters:
                    for converter in self.media_converters:
                        try:
                            # Check if the converter can convert the media:
                            if converter.can_convert(media):
                                # If the converter can convert the media, convert it:
                                converted = converter.convert(media)
                                # Break after the first successful conversion
                                break
                        except Exception as e:
                            aprint("Error during conversion:", e)
                            new_message.append_text(
                                f"Could not convert {type(media).__name__} to allowed types because of error {e}."
                            )
                            import traceback

                            traceback.print_exc()

                    if converted is not None:
                        conversion_happened = True
                        # If a converter was found, append the converted media
                        for converted_item in converted:
                            new_message.append_block(
                                MessageBlock(
                                    media=converted_item, attributes=attributes
                                )
                            )
                    else:
                        # If no converter is found, keep the original media:
                        new_message.append_block(block)

                        # if the original media is not allowed, log a warning:
                        if not any(
                            isinstance(media, allowed_type)
                            for allowed_type in allowed_media_types
                        ) and not isinstance(media, Action):
                            aprint(
                                f"Warning: No converter found for non-allowed media of type: {type(media)}"
                            )

            converted_messages.append(new_message)

        # If conversion happened, call convert recursively
        if recursive and conversion_happened:
            return self.convert(converted_messages, allowed_media_types, recursive=True)

        return converted_messages



import re

# Compile once on import
_GEMINI_REASONING_RE = re.compile(
    r"""
    ^(?:models/)?           # optional models/ prefix
    gemini-                 # prefix
    (?:1\.5|2\.0|2\.5|3)-   # supported generations (added 3)
    (?:                     # variations with reasoning
        pro |               # pro
        flash(?:-lite)? |   # flash or flash-lite
        flash-thinking      # flash-thinking experimental
    )
    (?:-.+)?$               # optional suffix: -preview-05-06, -exp-01-21, etc.
    """,
    re.VERBOSE,
)


def is_gemini_reasoning_model(model_name: str) -> bool:
    """
    True ⇔ `model_name` supports Gemini’s built-in thinking/reasoning trace.

    >>> is_gemini_reasoning_model("gemini-2.5-pro-preview-05-06")
    True
    >>> is_gemini_reasoning_model("gemini-nano")
    False
    """
    return bool(_GEMINI_REASONING_RE.fullmatch(model_name))



from typing import List, Optional

from pydantic import BaseModel

from litemind.agent.messages.actions.action_base import ActionBase
from litemind.agent.messages.actions.tool_call import ToolCall
from litemind.agent.messages.actions.tool_use import ToolUse
from litemind.agent.messages.message import Message
from litemind.media.types.media_action import Action
from litemind.media.types.media_document import Document
from litemind.media.types.media_image import Image
from litemind.media.types.media_text import Text


def convert_messages_for_anthropic(
    messages: List[Message],
    response_format: Optional[BaseModel] = None,
    cache_support: bool = True,
    media_converter: Optional["MediaConverter"] = None,
) -> List["MessageParam"]:
    """Convert litemind Messages into Anthropic's MessageParam format.

    Handles all block types including text, images, documents (PDF),
    tool calls/results, thinking blocks, and citations from built-in tools.

    Parameters
    ----------
    messages : List[Message]
        Litemind messages to convert.
    response_format : Optional[BaseModel]
        If provided, appends a JSON schema prompt to request structured output.
    cache_support : bool
        If True, adds ephemeral cache control to the last content block.
    media_converter : Optional[MediaConverter]
        Converter for non-PDF documents to markdown text.

    Returns
    -------
    List[MessageParam]
        Messages in Anthropic API format.
    """

    from anthropic.types import MessageParam

    anthropic_messages: List[MessageParam] = []

    for message in messages:
        content = []

        for block in message.blocks:
            if block.has_type(Text) and not block.is_thinking():

                # Get Text's string:
                text: str = block.get_content()

                if message.role == "assistant":
                    # remove trailing whitespace as it is not allowed by Anthropic!
                    text = text.rstrip()

                # Create text block
                text_block = {"type": "text", "text": text}

                # Handle citations if they exist in block attributes
                if (
                    hasattr(block, "attributes")
                    and block.attributes
                    and "citations" in block.attributes
                ):
                    citations = block.attributes["citations"]
                    if citations:
                        # Convert citations to Anthropic format
                        anthropic_citations = []
                        for citation in citations:
                            if citation.get("type") == "web_search_citation":
                                anthropic_citations.append(
                                    {
                                        "type": "web_search_result_location",
                                        "url": citation.get("url", ""),
                                        "title": citation.get("title", ""),
                                        "cited_text": citation.get("cited_text", ""),
                                    }
                                )

                        if anthropic_citations:
                            text_block["citations"] = anthropic_citations

                content.append(text_block)

            elif block.has_type(Text) and block.is_thinking():

                # Get Text's string:
                text: str = block.get_content()

                if block.is_redacted():
                    # if the block is redacted, add it as a redacted thinking block:
                    content.append({"type": "redacted_thinking", "data": text})
                else:
                    # Append the thinking text to the Anthropic's message content:
                    content.append(
                        {
                            "type": "thinking",
                            "thinking": text,
                            "signature": block.attributes.get("signature", ""),
                        }
                    )

            elif block.has_type(Image):

                # Cast to Image:
                image: Image = block.media

                # Check if we have a remote URL (http/https)
                if not image.is_local() and image.uri.startswith(
                    ("http://", "https://")
                ):
                    # Use URL-based source for remote images
                    content.append(
                        {
                            "type": "image",
                            "source": {
                                "type": "url",
                                "url": image.uri,
                            },
                        }
                    )
                else:
                    # Use base64-encoded source for local images
                    media_type = image.get_media_type()
                    base64_data = image.to_base64_data()
                    content.append(
                        {
                            "type": "image",
                            "source": {
                                "type": "base64",
                                "media_type": media_type,
                                "data": base64_data,
                            },
                        }
                    )

            elif block.has_type(Document):

                # Cast to Document:
                document: Document = block.media

                if document.has_extension("pdf"):
                    # Check if we have a remote URL (http/https)
                    if not document.is_local() and document.uri.startswith(
                        ("http://", "https://")
                    ):
                        # Use URL-based source for remote PDFs
                        content.append(
                            {
                                "type": "document",
                                "source": {
                                    "type": "url",
                                    "url": document.uri,
                                },
                            }
                        )
                    else:
                        # Use base64-encoded source for local PDFs
                        media_type = document.get_media_type()
                        base64_data = document.to_base64_data()
                        content.append(
                            {
                                "type": "document",
                                "source": {
                                    "type": "base64",
                                    "media_type": media_type,
                                    "data": base64_data,
                                },
                            }
                        )
                else:
                    # Convert Document to text:
                    markdown: str = document.to_markdown(
                        media_converter=media_converter
                    )

                    # Add the document text to the content:
                    content.append({"type": "text", "text": markdown})

            elif block.has_type(Action):

                # Get the tool action:
                tool_action: ActionBase = block.get_content()

                # if contents is a ToolUse object do the following:
                if isinstance(tool_action, ToolUse):

                    # Get the tool use object:
                    tool_use: ToolUse = tool_action

                    # Add the tool result to the content:
                    content.append(
                        {
                            "type": "tool_result",
                            "tool_use_id": tool_use.id,
                            "content": [
                                {
                                    "type": "text",
                                    "text": tool_use.result,
                                }
                            ],
                        }
                    )

                elif isinstance(tool_action, ToolCall):

                    # Get the tool call object:
                    tool_call: ToolCall = tool_action

                    # Add regular tool_use block
                    content.append(
                        {
                            "id": tool_call.id,
                            "type": "tool_use",
                            "name": tool_call.tool_name,
                            "input": tool_call.arguments,
                        }
                    )

                else:
                    raise ValueError(
                        f"Unsupported action type: {type(tool_action).__name__}"
                    )

            else:
                raise ValueError(
                    f"Unsupported block type: {type(block.media).__name__}"
                )

        anthropic_messages.append(
            {
                "role": message.role,
                "content": content,
            }
        )

    # Add cache_control to the last message if requested:
    if cache_support:
        if anthropic_messages and anthropic_messages[-1]["content"]:
            anthropic_messages[-1]["content"][-1]["cache_control"] = {
                "type": "ephemeral"
            }

    # Add prompt that specifies that response should be in JSON following given JSON schema:
    if response_format:
        anthropic_messages[-1]["content"].append(
            {
                "type": "text",
                "text": f"\nPlease provide answer as a JSON string that adheres to the following schema:\n{response_format.model_json_schema()}\n\n Without text before or after.",
            }
        )

    return anthropic_messages



from typing import List, Optional, Sequence, Union

from arbol import aprint, asection

from litemind.agent.augmentations.augmentation_base import AugmentationBase
from litemind.agent.augmentations.augmentation_set import AugmentationSet
from litemind.agent.augmentations.information.information_base import InformationBase
from litemind.agent.messages.conversation import Conversation
from litemind.agent.messages.message import Message
from litemind.agent.tools.base_tool import BaseTool
from litemind.agent.tools.toolset import ToolSet
from litemind.apis.base_api import BaseApi
from litemind.apis.model_features import ModelFeatures


class Agent:
    """Main orchestrator for agentic AI interactions.

    The Agent manages conversation state, tool execution, augmentation-based
    retrieval, and LLM API calls. It selects a suitable model based on
    requested features, maintains a conversation history, and supports
    Retrieval-Augmented Generation (RAG) via its augmentation set.
    """

    def __init__(
        self,
        api: BaseApi,
        name: str = "Agent",
        model_name: Optional[str] = None,
        model_features: Optional[
            Union[str, List[str], ModelFeatures, Sequence[ModelFeatures]]
        ] = None,
        temperature: float = 0.0,
        toolset: Optional[ToolSet] = None,
        augmentation_set: Optional[AugmentationSet] = None,
        augmentation_k: int = 5,
        augmentation_context_position: str = "before_query",
        messages_stdout_enabled: bool = False,
        **kwargs,
    ):
        """
        Create a new agent.

        Parameters
        ----------
        api: BaseApi
            The API to use.
        name: str
            The name of the agent.
        model_name: str
            The name of the model to use.
        model_features: Union[str, List[str], ModelFeatures, Sequence[ModelFeatures]]
            The features to require for a model. If features are provided, the
            model will be selected based on the features. A ValueError will be
            raised if a model has also been specified as this is mutually
            exclusive. Minimal required features are text generation and tools
            and are added automatically if not provided.
        temperature: float
            The temperature to use for text generation.
        toolset: ToolSet
            The toolset to use.
        augmentation_set: Optional[AugmentationSet]
            The set of augmentations to use for document retrieval.
        augmentation_k: int
            The number of informations to retrieve from augmentations.
        augmentation_context_position: str
            Where to place the retrieved context. Options are:
            - "before_query": Add context right before the user query
            - "system": Add context as a system message
        messages_stdout_enabled: bool
            If True, messages will be printed to stdout.
        kwargs: Any
            Additional keyword arguments to pass to the API.
        """

        # get key from environmental variables:
        self.api = api

        # Normalise the model features:
        model_features = ModelFeatures.normalise(model_features)

        # If model features are provided, get the best model based on the features:
        if model_features is not None:

            # Make sure that model is not also provided:
            if model_name:
                raise ValueError("You cannot provide both model and model_features")

            # Make sure that text generation and tools are included in the model features:
            if ModelFeatures.TextGeneration not in model_features:
                model_features = list(model_features) + [ModelFeatures.TextGeneration]
            if ModelFeatures.Tools not in model_features:
                model_features = list(model_features) + [ModelFeatures.Tools]

            # Get the best model based on the features:
            self.model = api.get_best_model(features=model_features)
        else:
            # Use the provided model or get the best model based on text generation and tools:
            self.model = model_name or api.get_best_model(
                features=[ModelFeatures.TextGeneration, ModelFeatures.Tools]
            )

        # Let's make sure the model is not None,
        if self.model is None:
            raise ValueError("No model was found with the required features")

        # ... and has the minimal required features:
        if not self.api.has_model_support_for(
            model_name=self.model, features=ModelFeatures.TextGeneration
        ):
            aprint(f"The model {self.model} does not seem to support text generation!")

        model_supports_tools = self.api.has_model_support_for(
            model_name=self.model, features=ModelFeatures.Tools
        )
        if toolset is not None and not model_supports_tools:
            aprint(f"The model {self.model} does not seem to support tools!")

        # Set the temperature, toolset, and name:
        self.temperature = temperature
        self.toolset = toolset or (ToolSet() if model_supports_tools else None)
        self.name = name
        self._agent_kwargs = kwargs

        # Initialize augmentation parameters
        self.augmentation_set = augmentation_set or AugmentationSet()
        self.augmentation_k = augmentation_k

        # Validate context_position
        valid_positions = ["before_query", "system"]
        if augmentation_context_position not in valid_positions:
            raise ValueError(
                f"Invalid augmentation_context_position: {augmentation_context_position}. "
                f"Valid options are: {valid_positions}"
            )
        self.augmentation_context_position = augmentation_context_position

        # Initialise conversation:
        self.conversation = Conversation()

        self.messages_stdout_enabled = messages_stdout_enabled

    def append_system_message(self, message: Union[str, Message]):
        """
        Append a system message to the agent's conversation.

        Parameters
        ----------
        message : Union[str, Message]
            The system message to append. If a string is provided, it is
            automatically wrapped in a Message with role='system'.
        """

        # If a string is provided, convert it to a message:
        if isinstance(message, str):
            message = Message(role="system", text=message)

        # Append the message to the conversation:
        self.conversation.system_messages.append(message)

    def add_tool(self, tool: BaseTool) -> None:
        """
        Add a tool to the agent's toolset.

        Parameters
        ----------
        tool: BaseTool
            The tool to add.
        """
        if self.toolset is None:
            raise ValueError("The model does not support tools")
        self.toolset.add_tool(tool)

    def remove_tool(self, name: str) -> bool:
        """
        Remove a tool by name from the agent's toolset.

        Parameters
        ----------
        name: str
            The name of the tool to remove.

        Returns
        -------
        bool
            True if the tool was removed, False if not found.
        """
        if self.toolset is None:
            raise ValueError("The model does not support tools")
        return self.toolset.remove_tool(name)

    def clear_tools(self) -> None:
        """Clear all tools from the agent's toolset."""
        if self.toolset is None:
            raise ValueError("The model does not support tools")
        self.toolset = ToolSet()

    def add_augmentation(
        self,
        augmentation: AugmentationBase,
        k: int = 5,
        threshold: float = 0.0,
    ) -> None:
        """
        Add an augmentation to the agent's augmentation set.

        Parameters
        ----------
        augmentation: AugmentationBase
            The augmentation to add.
        k: int
            The number of informations to retrieve from the augmentation.
        threshold: float
            The score threshold for the augmentation. If the score is below
            this value, the augmentation will not be used.
        """
        self.augmentation_set.add_augmentation(augmentation, k=k, threshold=threshold)

    def remove_augmentation(self, name: str) -> bool:
        """
        Remove an augmentation by name from the agent's augmentation set.

        Parameters
        ----------
        name: str
            The name of the augmentation to remove.

        Returns
        -------
        bool
            True if the augmentation was removed, False if not found.
        """
        return self.augmentation_set.remove_augmentation(name)

    def clear_augmentations(self) -> None:
        """Clear all augmentations from the agent's augmentation set."""
        self.augmentation_set = AugmentationSet()

    def list_augmentations(self) -> List[AugmentationBase]:
        """
        Get all augmentations in the agent's augmentation set.

        Returns
        -------
        List[AugmentationBase]
            The list of augmentations.
        """
        return self.augmentation_set.list_augmentations()

    def _retrieve_relevant_documents(self, query_text: str) -> List[InformationBase]:
        """
        Retrieve relevant informations from the augmentation set.

        Parameters
        ----------
        query_text: str
            The query to retrieve informations for.

        Returns
        -------
        List[InformationBase]
            A list of relevant informations.
        """
        with asection(
            f"Retrieving relevant informations for query: '{query_text[:100]}...'"
        ):
            documents = self.augmentation_set.search_combined(
                query_text, k=self.augmentation_k
            )
            with asection(
                f"Retrieved {len(documents)} informations from augmentation set:"
            ):
                for doc in documents:
                    aprint(f"Document ID: {doc.id}, Score: {doc.score:.4f}")
            return documents

    def _add_context_to_conversation(
        self,
        query_message: Message,
        documents: List[InformationBase],
        max_doc_length: int = 1000,
    ) -> None:
        """
        Add retrieved context from informations to the conversation.

        Depending on ``augmentation_context_position``, the context is
        inserted either as a user message right before the query or as an
        additional system message.

        Parameters
        ----------
        query_message : Message
            The user's query message.
        documents : List[InformationBase]
            The informations to add as context.
        max_doc_length : int
            Maximum character length for each document's context text.
        """
        if not documents:
            return

        # Add the context based on the specified position
        if self.augmentation_context_position == "before_query":
            # Create a user message with the context
            augmentation_message = Message(role="user")

            # Insert the context message right before the user query
            idx = self.conversation.standard_messages.index(query_message)
            self.conversation.standard_messages.insert(idx, augmentation_message)

        elif self.augmentation_context_position == "system":
            # Add as a system message
            augmentation_message = Message(role="system")
            self.conversation.system_messages.append(augmentation_message)

        else:
            raise ValueError(
                f"Invalid augmentation_context_position: {self.augmentation_context_position}"
            )

        augmentation_message.append_text("Additional context information:\n\n")

        for i, doc in enumerate(documents):

            # prepare the context text
            context_text = ""
            context_text += f"--- Document {i + 1} "
            if doc.score is not None:
                context_text += f"(Relevance: {doc.score:.4f}) "
            if "augmentation" in doc.metadata:
                context_text += f"[Source: {doc.metadata['augmentation']}] "
            context_text += "---\n"

            # Add document context:
            augmentation_message.append_text(context_text)

            # Add document content:
            augmentation_message += doc.to_message_block()

        with asection(
            f"Added {len(documents)} informations to conversation, context message:"
        ):
            aprint(str(augmentation_message))

    def __getitem__(self, item) -> Message:
        return self.conversation[item]

    def __len__(self):
        return len(self.conversation)

    def __iadd__(self, other: Message):
        self.conversation.append(other)
        return self

    def __call__(self, *args, **kwargs) -> List[Message]:
        """
        Run the agent on the given input and return the response messages.

        Accepts text strings, Message objects, lists of messages, or
        Conversation objects as positional or keyword arguments. The input
        is appended to the conversation, augmentations are queried if
        available, and the LLM API is called (with tool execution if
        applicable).

        Parameters
        ----------
        *args
            Positional arguments: str (user text), Message, list of Messages,
            or a Conversation.
        **kwargs
            Keyword arguments including ``text``, ``message``,
            ``conversation``, and ``role``.

        Returns
        -------
        List[Message]
            The response messages from the LLM, which are also appended
            to the conversation history.
        """

        # Prepare the call by normalising parameters:
        self._prepare_call(*args, **kwargs)

        # Get last message in conversation
        last_message = self.conversation.get_last_message()

        with asection(f"Calling agent: '{self.name}'"):

            with asection("API and model:"):
                aprint(f"API: {self.api.__class__.__name__}")
                aprint(f"Model: {self.model}")

            if self.toolset:
                with asection("Available tools"):
                    if len(self.toolset) > 0:
                        for tool in self.toolset:
                            aprint(tool.pretty_string())
                    else:
                        aprint("No tools available")

            with asection("Available augmentations"):
                augmentations = self.augmentation_set.list_augmentations()
                if augmentations:
                    for aug in augmentations:
                        aprint(f"{aug.name}")
                else:
                    aprint("No augmentations available")

            # Quick report of the conversation (just number of messages):
            if self.conversation:
                with asection("Conversation report:"):
                    aprint(
                        f"Number of messages currently in conversation: {len(self.conversation)}"
                    )

            # Print the last message in conversation if enabled
            if self.messages_stdout_enabled:
                with asection("Last message in conversation:"):
                    aprint(last_message)

            # Process augmentations if we have any and there's a query message
            if len(self.augmentation_set) > 0 and last_message:
                query_text = last_message.to_plain_text()

                # Retrieve relevant informations
                relevant_documents = self._retrieve_relevant_documents(query_text)

                # Add context to the conversation
                if relevant_documents:
                    self._add_context_to_conversation(last_message, relevant_documents)

            # Call the API:
            response = self.api.generate_text(
                model_name=self.model,
                messages=self.conversation.get_all_messages(),
                temperature=self.temperature,
                toolset=self.toolset,
                **kwargs,
            )

            # Quick report of the response:
            with asection("Response:"):
                if response is None:
                    aprint(
                        "No response received from the API. Probably an error occurred."
                    )
                else:
                    aprint(f"Number of messages in response: {len(response)}")

            # Print response messages to stdout if enabled:
            if self.messages_stdout_enabled and response:
                with asection("Response:"):
                    for message in response:
                        aprint(message)

        # Append response messages to conversation:
        if response:
            self.conversation.extend(response)

        # Return response
        return response

    def _prepare_call(self, *args, **kwargs) -> List[Message]:
        """
        Normalize and append input arguments to the conversation.

        Handles multiple input formats (str, Message, list of Messages,
        Conversation) provided as positional or keyword arguments, and
        appends them to the agent's conversation.

        Parameters
        ----------
        *args
            Positional arguments: str, Message, list of Messages, or
            a Conversation.
        **kwargs
            Keyword arguments: ``text``, ``message``, ``conversation``,
            ``role``.

        Returns
        -------
        List[Message]
            The messages that were appended to the conversation.
        """

        messages: List[Message] = []

        # if a role is provided as a keyword argument, append it:
        role = kwargs.pop("role", "user")
        # if conversation is provided as a positional argument, append it:
        if args:
            conversation = args[0]
            if isinstance(conversation, Conversation):
                self.conversation += conversation
                messages.extend(conversation.get_all_messages())

        # if a list of messages is provided as a positional argument, append them:
        if args:
            first_arg = args[0]
            # Check for list/tuple of messages (use duck typing for RPyC compatibility)
            if isinstance(first_arg, (list, tuple)) and not isinstance(
                first_arg, Conversation
            ):
                for item in first_arg:
                    # Duck typing: check for 'role' attribute (Message-like)
                    if isinstance(item, Message) or hasattr(item, "role"):
                        self.conversation.append(item)
                        messages.append(item)

        # if conversation is provided as a keyword argument, append it:
        if "conversation" in kwargs:
            conversation = kwargs["conversation"]
            # remove conversation from kwargs:
            del kwargs["conversation"]
            if isinstance(conversation, Conversation):
                self.conversation += conversation
                messages.extend(conversation.get_all_messages())
            else:
                raise ValueError("conversation must be a Conversation object")

        # if a message is provided as a positional argument, append it:
        if args:
            for message in args:
                if isinstance(message, Message):
                    self.conversation.append(message)
                    messages.append(message)

        # if a message is provided as a keyword argument, append it:
        if "message" in kwargs:
            message = kwargs["message"]
            # remove message from kwargs:
            del kwargs["message"]
            if isinstance(message, Message):
                self.conversation.append(message)
                messages.append(message)
            else:
                raise ValueError("message must be a Message object")

        # if text is provided as a positional argument, append it:
        if args:
            for text in args:
                if isinstance(text, str):
                    message = Message(role=role)
                    message.append_text(text)
                    self.conversation.append(message)
                    messages.append(message)

        # if text is provided as a keyword argument, append it:
        if "text" in kwargs:
            text = kwargs["text"]
            # remove text from kwargs:
            del kwargs["text"]
            if isinstance(text, str):
                message = Message(role=role, text=text)
                self.conversation.append(message)
                messages.append(message)
            else:
                raise ValueError("text must be a string")

        return messages

    def __repr__(self):
        return f"{self.name}(model={self.model}, api={self.api.__class__.__name__})"

    def __str__(self):
        return self.__repr__()



import datetime
import inspect
import json
from typing import Any, Dict, Optional


class JSONSerializable:
    """
    Mixin class providing automatic JSON serialization.

    Supports recursive handling of nested ``JSONSerializable`` objects,
    lists, and dictionaries. Skips private attributes (prefixed with ``_``).
    """

    def to_json(self, indent: Optional[int] = None) -> str:
        """
        Serialize this object to a JSON string.

        Parameters
        ----------
        indent : int, optional
            Number of spaces for JSON indentation. None for compact output.

        Returns
        -------
        str
            The JSON string representation.
        """
        return json.dumps(self.to_dict(), cls=_JSONEncoder, indent=indent)

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert this object to a dictionary.

        Recursively converts nested ``JSONSerializable`` objects. Skips
        private attributes (those starting with ``_``).

        Returns
        -------
        dict
            Dictionary representation of this object.
        """
        result = {}
        for k, v in self.__dict__.items():
            if k.startswith("_"):
                continue

            # Handle nested JSONSerializable objects
            if isinstance(v, JSONSerializable):
                result[k] = v.to_dict()
            # Handle lists of objects
            elif isinstance(v, list):
                result[k] = [
                    item.to_dict() if isinstance(item, JSONSerializable) else item
                    for item in v
                ]
            # Handle dictionaries
            elif isinstance(v, dict):
                result[k] = {
                    key: (
                        value.to_dict()
                        if isinstance(value, JSONSerializable)
                        else value
                    )
                    for key, value in v.items()
                }
            else:
                result[k] = v
        return result

    @classmethod
    def from_json(cls, json_str: str) -> "JSONSerializable":
        """
        Create an instance from a JSON string.

        Parameters
        ----------
        json_str : str
            A valid JSON string.

        Returns
        -------
        JSONSerializable
            A new instance populated from the JSON data.

        Raises
        ------
        ValueError
            If the string is not valid JSON.
        """
        try:
            data = json.loads(json_str)
            return cls.from_dict(data)
        except json.JSONDecodeError:
            raise ValueError("Invalid JSON string")

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "JSONSerializable":
        """
        Create an instance from a dictionary.

        Matches dictionary keys to constructor parameters and sets any
        remaining keys as attributes on the instance.

        Parameters
        ----------
        data : dict
            Dictionary with keys corresponding to the class attributes.

        Returns
        -------
        JSONSerializable
            A new instance populated from the dictionary.
        """
        # Extract constructor parameters
        init_signature = inspect.signature(cls.__init__)
        param_names = set(list(init_signature.parameters.keys())[1:])  # Exclude 'self'

        # Build constructor arguments
        kwargs = {}
        for param in param_names:
            if param in data:
                value = data[param]

                # Special handling for nested objects
                if param == "child" and isinstance(value, dict) and "name" in value:
                    kwargs[param] = cls.from_dict(value)
                else:
                    kwargs[param] = value

        # Create instance
        instance = cls(**kwargs)

        # Set remaining attributes (that weren't passed to constructor)
        for k, v in data.items():
            if k not in kwargs:
                setattr(instance, k, v)

        return instance


class _JSONEncoder(json.JSONEncoder):
    """Custom JSON encoder that handles dates, sets, and objects with ``to_dict``."""

    def default(self, obj: Any) -> Any:
        # Handle dates and times
        if isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.isoformat()

        # Handle sets
        if isinstance(obj, set):
            return list(obj)

        # Handle objects with to_dict method
        if hasattr(obj, "to_dict") and callable(obj.to_dict):
            return obj.to_dict()

        return super().default(obj)



import os
from typing import List, Optional, Sequence, Set, Type, Union

from PIL import Image as PilImage
from pydantic import BaseModel

from litemind.agent.messages.message import Message
from litemind.agent.tools.toolset import ToolSet
from litemind.apis.base_api import ModelFeatures
from litemind.apis.callbacks.api_callback_manager import ApiCallbackManager
from litemind.apis.default_api import DefaultApi
from litemind.apis.exceptions import APIError, APINotAvailableError
from litemind.apis.model_registry import get_default_model_registry
from litemind.apis.providers.google.utils.aggregate_chat_response import (
    aggregate_chat_response,
)
from litemind.apis.providers.google.utils.check_availability import (
    check_gemini_api_availability,
)
from litemind.apis.providers.google.utils.convert_messages import (
    convert_messages_for_gemini,
)
from litemind.apis.providers.google.utils.format_tools import format_tools_for_gemini
from litemind.apis.providers.google.utils.list_models import _get_gemini_models_list
from litemind.apis.providers.google.utils.process_response import (
    process_response_from_gemini,
)
from litemind.media.media_base import MediaBase
from litemind.media.types.media_action import Action
from litemind.media.types.media_text import Text

# Default token limits for unknown models
_DEFAULT_CONTEXT_WINDOW = 1_000_000
_DEFAULT_MAX_OUTPUT_TOKENS = 65_536

# Supported aspect ratios for Gemini image generation
_SUPPORTED_ASPECT_RATIOS = {
    1.0: "1:1",
    0.75: "3:4",
    1.333: "4:3",
    0.5625: "9:16",
    1.778: "16:9",
}


def _snap_to_aspect_ratio(aspect_ratio: float) -> str:
    """Snap to closest supported Gemini aspect ratio."""
    closest = min(_SUPPORTED_ASPECT_RATIOS.keys(), key=lambda r: abs(r - aspect_ratio))
    return _SUPPORTED_ASPECT_RATIOS[closest]


class GeminiApi(DefaultApi):
    """
    A Gemini 1.5+ API implementation conforming to the BaseApi interface.
    Uses the google.genai library (unified Google GenAI SDK).

    Gemini models support text generation, image inputs, audio and video inputs, and thinking,
    but do not support all features natively. Support for these features
    are provided by the Litemind API via our fallback mechanism.

    Set the GOOGLE_GEMINI_API_KEY environment variable to your Google API key.
    You can get a key from the Google Cloud Console: https://ai.google.dev/gemini-api/docs/api-key

    Note: The API key must be enabled for the Gemini API.
    You can also pass the API key explicitly to the constructor.

    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        allow_media_conversions: bool = True,
        allow_media_conversions_with_models: bool = True,
        callback_manager: Optional[ApiCallbackManager] = None,
        **kwargs,
    ):
        """
        Initialize the Gemini client.

        Parameters
        ----------
        api_key : Optional[str]
            The API key for Google GenAI. If not provided, reads from GOOGLE_GEMINI_API_KEY.
        allow_media_conversions: bool
            If True, the API will allow media conversions using the default media converter.
        allow_media_conversions_with_models: bool
            If True, the API will allow media conversions using models that support
            the required features in addition to the default media converter.
            To use this, the allow_media_conversions parameter must be True.
        callback_manager : Optional[ApiCallbackManager]
            Callback manager for API events.
        kwargs : dict
            Additional parameters (unused here, but accepted for consistency).
        """

        super().__init__(
            allow_media_conversions=allow_media_conversions,
            allow_media_conversions_with_models=allow_media_conversions_with_models,
            callback_manager=callback_manager,
        )

        if api_key is None:
            api_key = os.environ.get("GOOGLE_GEMINI_API_KEY")
        if not api_key:
            raise APIError(
                "A valid GOOGLE_GEMINI_API_KEY is required for GeminiApi. "
                "Set GOOGLE_GEMINI_API_KEY in the environment or pass api_key explicitly."
            )

        self._api_key = api_key
        self.kwargs = kwargs

        try:
            # Initialize the model registry for feature lookups:
            self.model_registry = get_default_model_registry()

            # google.genai references - new unified SDK
            from google.genai import Client

            # Create the client instance (replaces global genai.configure)
            # Note: Using default timeout as custom timeouts can cause issues
            self.client = Client(api_key=self._api_key)

            # Get the raw model list:
            self._model_list = _get_gemini_models_list(self.client)

        except Exception as e:
            # Print stack trace:
            import traceback

            traceback.print_exc()
            raise APINotAvailableError(f"Error initializing Gemini client: {e}")

    def check_availability_and_credentials(self, api_key: Optional[str] = None) -> bool:

        # Check the availability of the API using the client:
        result = check_gemini_api_availability(self.client)

        # Call the callback manager:
        self.callback_manager.on_availability_check(result)

        # Return the result:
        return result

    def list_models(
        self,
        features: Optional[
            Union[str, List[str], ModelFeatures, Sequence[ModelFeatures]]
        ] = None,
        non_features: Optional[
            Union[str, List[str], ModelFeatures, Sequence[ModelFeatures]]
        ] = None,
        media_types: Optional[Set[Type[MediaBase]]] = None,
    ) -> List[str]:

        try:
            # Normalise the features:
            features = ModelFeatures.normalise(features)
            non_features = ModelFeatures.normalise(non_features)

            # Get the full list of models:
            model_list = list(self._model_list)

            # Add the models from the super class:
            model_list += super().list_models()

            # To be safe, we remove duplicates but keep the order:
            model_list = list(dict.fromkeys(model_list))

            # Filter the models based on the features:
            if features:
                model_list = self._filter_models(
                    model_list,
                    features=features,
                    non_features=non_features,
                    media_types=media_types,
                )

            # Call _callbacks:
            self.callback_manager.on_model_list(model_list)

            return model_list

        except Exception:
            raise APIError("Error fetching model list from Google.")

    def get_best_model(
        self,
        features: Optional[
            Union[str, List[str], ModelFeatures, Sequence[ModelFeatures]]
        ] = None,
        non_features: Optional[
            Union[str, List[str], ModelFeatures, Sequence[ModelFeatures]]
        ] = None,
        media_types: Optional[Sequence[Type[MediaBase]]] = None,
        exclusion_filters: Optional[Union[str, List[str]]] = None,
    ) -> Optional[str]:

        # Normalise the features:
        features = ModelFeatures.normalise(features)
        non_features = ModelFeatures.normalise(non_features)

        # Get the full list of models:
        model_list = self.list_models()

        # Filter the models based on the requirements:
        model_list = self._filter_models(
            model_list,
            features=features,
            non_features=non_features,
            media_types=media_types,
            exclusion_filters=exclusion_filters,
        )

        if model_list:
            # If we have any models left, return the first one:
            model_name = model_list[0]
        else:
            model_name = None

        # Call the _callbacks:
        self.callback_manager.on_best_model_selected(model_name)

        return model_name

    def has_model_support_for(
        self,
        features: Union[str, List[str], ModelFeatures, Sequence[ModelFeatures]],
        media_types: Optional[Sequence[Type[MediaBase]]] = None,
        model_name: Optional[str] = None,
    ) -> bool:

        # Normalise the features:
        features = ModelFeatures.normalise(features)

        # Get the best model if not provided:
        if model_name is None:
            model_name = self.get_best_model(features=features, media_types=media_types)

        # If model_name is None then we return False:
        if model_name is None:
            return False

        # We check if the superclass says that the model supports the features:
        if super().has_model_support_for(
            features=features, media_types=media_types, model_name=model_name
        ):
            return True

        for feature in features:
            if not self.model_registry.supports_feature(
                self.__class__, model_name, feature
            ):
                # If the model does not support the feature, we return False:
                return False

        return True

    def _has_thinking_support(self, model_name: str) -> bool:
        """Check if model supports thinking via registry."""
        return self.model_registry.supports_feature(
            self.__class__, model_name, ModelFeatures.Thinking
        )

    def _build_thinking_config(self, model_name: str, max_output_tokens: int):
        """
        Build a ThinkingConfig for models that support native thinking.

        Parameters
        ----------
        model_name : str
            The model name to check for thinking support.
        max_output_tokens : int
            Maximum output tokens to compute thinking budget.

        Returns
        -------
        ThinkingConfig or None
            ThinkingConfig if the model supports thinking, None otherwise.
        """
        if not self._has_thinking_support(model_name):
            return None

        from google.genai import types

        model_lower = model_name.lower()

        # Gemini 3.x uses thinking_level instead of thinking_budget (per Google docs)
        # - "low" may not consistently return thinking content
        # - "medium" is only supported by Gemini 3 Flash, not Pro
        # - "high" is supported by all Gemini 3 models and consistently returns thinking
        # include_thoughts=True is required to get thinking content in the response
        if "gemini-3" in model_lower:
            # Use "high" for consistent thinking output across Gemini 3 Pro and Flash
            return types.ThinkingConfig(
                thinking_level="high",
                include_thoughts=True,
            )
        else:
            # Gemini 2.5 and other thinking models use thinking_budget
            thinking_budget = max(1024, max_output_tokens // 3)
            return types.ThinkingConfig(
                thinking_budget=thinking_budget,
                include_thoughts=True,
            )

    def max_num_input_tokens(self, model_name: Optional[str] = None) -> int:
        if model_name is None:
            model_name = self.get_best_model()
        model_info = self.model_registry.get_model_info(self.__class__, model_name)
        if model_info and model_info.context_window:
            return model_info.context_window
        return _DEFAULT_CONTEXT_WINDOW

    def max_num_output_tokens(self, model_name: Optional[str] = None) -> int:
        if model_name is None:
            model_name = self.get_best_model()
        model_info = self.model_registry.get_model_info(self.__class__, model_name)
        if model_info and model_info.max_output_tokens:
            return model_info.max_output_tokens
        return _DEFAULT_MAX_OUTPUT_TOKENS

    def generate_text(
        self,
        messages: List[Message],
        model_name: Optional[str] = None,
        temperature: float = 0.0,
        max_num_output_tokens: Optional[int] = None,
        toolset: Optional[ToolSet] = None,
        use_tools: bool = True,
        response_format: Optional[BaseModel] = None,
        **kwargs,
    ) -> List[Message]:

        from google.genai import types

        # validate inputs:
        super().generate_text(
            messages=messages,
            model_name=model_name,
            temperature=temperature,
            max_num_output_tokens=max_num_output_tokens,
            toolset=toolset,
            use_tools=use_tools,
            response_format=response_format,
            **kwargs,
        )

        # Set default model if not provided
        if model_name is None:
            model_name = self._get_best_model_for_text_generation(
                messages, toolset if use_tools else None, response_format
            )

        # Get system instruction from messages:
        system_instruction = ""
        for message in messages:
            if message.role == "system":
                for block in message.blocks:
                    if block.has_type(Text):
                        system_instruction += block.get_content()

        # Convert user messages -> gemini format
        preprocessed_messages = self._preprocess_messages(
            messages=messages,
            allowed_media_types=self._get_allowed_media_types_for_text_generation(
                model_name=model_name
            ),
        )

        # Get max num of output tokens for model if not provided:
        if max_num_output_tokens is None:
            max_num_output_tokens = self.max_num_output_tokens(model_name)

        # Build native ThinkingConfig if model supports it
        thinking_config = self._build_thinking_config(model_name, max_num_output_tokens)

        # Format tools for Gemini:
        gemini_tools = format_tools_for_gemini(toolset)

        # Disable explicit thinking config when tools are present — Gemini
        # models that support thinking will still think implicitly, but we
        # avoid passing a thinking_config that could conflict with tool use.
        if gemini_tools:
            thinking_config = None

        # Build GenerationConfig (tools are part of the config in new SDK)
        if response_format is None or toolset is not None:
            generation_cfg = types.GenerateContentConfig(
                temperature=temperature,
                max_output_tokens=max_num_output_tokens,
                system_instruction=system_instruction if system_instruction else None,
                thinking_config=thinking_config,
                tools=gemini_tools,
            )
        else:
            generation_cfg = types.GenerateContentConfig(
                temperature=temperature,
                max_output_tokens=max_num_output_tokens,
                response_mime_type="application/json",
                response_schema=response_format,
                system_instruction=system_instruction if system_instruction else None,
                thinking_config=thinking_config,
                tools=gemini_tools,
            )

        # List of new messages part of the response:
        new_messages = []

        try:

            # Loop until we get a response that doesn't require tool use:
            while True:

                # Convert messages to gemini format:
                gemini_messages = convert_messages_for_gemini(
                    preprocessed_messages, self.client
                )

                # Send messages to Gemini using streaming:
                streaming_response = self.client.models.generate_content_stream(
                    model=model_name,
                    contents=gemini_messages,
                    config=generation_cfg,
                )

                # Aggregate the response:
                gemini_response = aggregate_chat_response(
                    streaming_response, self.callback_manager.on_text_streaming
                )

                # Process the response:
                response = process_response_from_gemini(
                    gemini_response,
                    response_format=response_format,
                )

                # Append response message to original, preprocessed, and new messages:
                messages.append(response)
                preprocessed_messages.append(response)
                new_messages.append(response)

                # If the model wants to use a tool, parse out the tool calls:
                if not response.has(Action):
                    # Break out of the loop if no tool use is required anymore:
                    break

                if not use_tools:
                    # Break out of the loop if we're not using tools:
                    break

                # Process the tool call
                self._process_tool_calls(
                    response,
                    messages,
                    new_messages,
                    preprocessed_messages,
                    toolset,
                    set_preprocessed=False,
                )

            # Fire final callback:
            kwargs.update(
                {
                    "model_name": model_name,
                    "temperature": temperature,
                    "max_output_tokens": max_num_output_tokens,
                    "toolset": toolset,
                    "response_format": response_format,
                }
            )

            # Call the callback manager:
            self.callback_manager.on_text_generation(
                response=response, messages=messages, **kwargs
            )

        except Exception as e:
            raise APIError(f"Gemini generate text error: {e}")

        return new_messages

    def generate_image(
        self,
        positive_prompt: str,
        negative_prompt: Optional[str] = None,
        model_name: str = None,
        image_width: int = 512,
        image_height: int = 512,
        preserve_aspect_ratio: bool = True,
        allow_resizing: bool = True,
        **kwargs,
    ) -> PilImage:

        if model_name is None:
            model_name = self.get_best_model(features=ModelFeatures.ImageGeneration)

        from google.genai import types

        # Computes aspect ratio from resolution and snap to closest supported:
        aspect_ratio = image_width / image_height
        aspect_ratio_str = _snap_to_aspect_ratio(aspect_ratio)

        # Build the generation config
        # Note: Imagen models only support "BLOCK_LOW_AND_ABOVE" safety filter level
        config = types.GenerateImagesConfig(
            number_of_images=1,
            safety_filter_level="BLOCK_LOW_AND_ABOVE",
            person_generation="ALLOW_ADULT",
            aspect_ratio=aspect_ratio_str,
            negative_prompt=negative_prompt,
        )

        # Generate images using the client
        result = self.client.models.generate_images(
            model=model_name,
            prompt=positive_prompt,
            config=config,
        )

        # Get the generated image and ensure it is a PIL Image:
        google_image = result.generated_images[0].image

        # The Google SDK may return a google.genai.types.Image wrapper
        # rather than a PIL Image directly. Convert if needed:
        if isinstance(google_image, PilImage.Image):
            generated_image = google_image
        elif hasattr(google_image, "_pil_image") and google_image._pil_image:
            generated_image = google_image._pil_image
        elif hasattr(google_image, "image_bytes") and google_image.image_bytes:
            import io

            generated_image = PilImage.open(io.BytesIO(google_image.image_bytes))
        else:
            raise ValueError(
                "Could not extract PIL image from Gemini generate_images response."
            )

        # Add all parameters to kwargs:
        kwargs.update(
            {
                "negative_prompt": negative_prompt,
                "model_name": model_name,
                "image_width": image_width,
                "image_height": image_height,
                "preserve_aspect_ratio": preserve_aspect_ratio,
                "allow_resizing": allow_resizing,
            }
        )

        # Call the callback manager:
        self.callback_manager.on_image_generation(
            prompt=positive_prompt, image=generated_image, **kwargs
        )

        return generated_image

    def embed_texts(
        self,
        texts: Sequence[str],
        model_name: Optional[str] = None,
        dimensions: int = 512,
        **kwargs,
    ) -> Sequence[Sequence[float]]:

        # Get the best model if not provided:
        if model_name is None:
            model_name = self.get_best_model(features=ModelFeatures.TextEmbeddings)

        # If model belongs to super class delegate to it:
        if model_name in super().list_models():
            return super().embed_texts(
                texts=texts, model_name=model_name, dimensions=dimensions, **kwargs
            )

        from google.genai import types

        # Generate the embeddings:
        embeddings = []

        for text in texts:
            # Build embedding config
            config = types.EmbedContentConfig(
                output_dimensionality=dimensions,
            )

            # Generate the embeddings using the client:
            result = self.client.models.embed_content(
                model=model_name,
                contents=text,
                config=config,
            )

            # Get the embeddings:
            embedding = result.embeddings[0].values

            # Get the embeddings:
            embeddings.append(embedding)

        # Add other parameters to kwargs dict:
        kwargs.update({"model_name": model_name, "dimensions": dimensions})

        # Call the callback manager:
        self.callback_manager.on_text_embedding(
            texts=texts, embeddings=embeddings, **kwargs
        )

        return embeddings



from typing import List, Tuple, Type

from litemind.media.conversion.converters.base_converter import BaseConverter
from litemind.media.media_base import MediaBase
from litemind.media.types.media_audio import Audio
from litemind.media.types.media_text import Text
from litemind.utils.normalise_uri_to_local_file_path import uri_to_local_file_path
from litemind.utils.whisper_transcribe_audio import (
    is_local_whisper_available,
    transcribe_audio_with_local_whisper,
)


class AudioConverterWhisperLocal(BaseConverter):
    """Converts Audio media to Text using a local Whisper model.

    Produces a Text media containing audio metadata and the transcription.
    Requires the Whisper library to be installed locally.
    """

    def rule(self) -> List[Tuple[Type[MediaBase], List[Type[MediaBase]]]]:
        return [(Audio, [Text])]

    def can_convert(self, media: MediaBase) -> bool:

        return media is not None and isinstance(media, Audio)

    def convert(self, media: MediaBase) -> List[MediaBase]:

        if not isinstance(media, Audio):
            raise ValueError(f"Expected Audio media, got {type(media)}")

        # Check if Whisper is available:
        if not is_local_whisper_available():
            raise RuntimeError(
                "Whisper is not available. Please install Whisper to use this converter."
            )

        # Get the media URI:
        media_uri = media.uri

        # Convert the document URI to a local file path.
        audio_path = uri_to_local_file_path(media_uri)

        # Get metadata from the audio file:
        audio_info = media.get_info_markdown()

        # Transcribe the audio file:
        transcription = transcribe_audio_with_local_whisper(audio_path)

        media_list = []

        media_list.append(
            Text(
                audio_info
                + "\n\nThe following is the transcription of the audio:\n"
                + transcription
            )
        )

        return media_list



from typing import Tuple


def find_free_port() -> Tuple[str, int]:
    """Finds and returns a free port and its associated host.

    This utility function temporarily binds a socket to an ephemeral port
    chosen by the operating system (by specifying port 0). It then
    retrieves the assigned port number, closes the socket to release it,
    and returns the port number. This is useful for running tests or services
    on a dynamically assigned port to avoid conflicts.

    Returns
    -------
    Tuple[str, int]
        A tuple containing the host address (str) and a free port number (int).

    """
    # Import Python's built-in library for low-level network operations.
    import socket

    # 1. Create a new socket object.
    # AF_INET specifies the address family for IPv4.
    # SOCK_STREAM specifies the socket type for TCP (as opposed to UDP).
    sock = socket.socket()

    # 2. Set the SO_REUSEADDR option.
    # This is a crucial step for test servers. It allows the kernel to reuse
    # a local socket in TIME_WAIT state, which can happen if the test server
    # is stopped and restarted quickly. This prevents "Address already in
    # use" errors during rapid test cycles.
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    # 3. Bind the socket to an address. This is the key step.
    # The first argument, '', is a shorthand for '0.0.0.0', meaning the socket
    # will be accessible from any of the machine's network interfaces.
    # The second argument, 0, is the special port number that tells the
    # operating system to find and assign any currently available ephemeral port.
    sock.bind(("", 0))

    # 4. Retrieve the assigned address and port.
    # Now that the OS has assigned a port, getsockname() returns a tuple
    # containing the host IP and the dynamically assigned port number.
    host, port = sock.getsockname()

    # 5. Close the socket.
    # This immediately releases the port, making it available for our actual
    # server to use right away. The OS is unlikely to assign this same port
    # to another process in the tiny fraction of a second before our test
    # server grabs it.
    sock.close()

    # The 'port' variable now holds the number of a port that was free
    # just moments ago, making it safe to use for the test server.
    return host, port



from litemind.media.media_default import MediaDefault


class Text(MediaDefault):
    """Media that stores plain text content.

    This is the most fundamental media type and serves as the common target
    for most media converters.
    """

    def __init__(self, text: str, **kwargs):
        """Create a new text media.

        Parameters
        ----------
        text : str
            The plain text content.
        **kwargs
            Additional keyword arguments forwarded to ``MediaDefault``.

        Raises
        ------
        ValueError
            If *text* is not a string.
        """

        super().__init__(**kwargs)

        # Check that it is really a string:
        if not isinstance(text, str):
            raise ValueError(f"Text must be a string, not {type(text)}")

        # Set attributes:
        self.text = text

    def get_content(self) -> str:
        """Return the plain text content.

        Returns
        -------
        str
            The stored text string.
        """
        return self.text

    def __str__(self) -> str:
        return self.text



from typing import Any, List, Optional, Union

from anthropic.types import WebSearchResultBlock
from pydantic import BaseModel

from litemind.agent.messages.message import Message
from litemind.utils.json_to_object import json_to_object


def process_response_from_anthropic(
    anthropic_response: Any,
    response_format: Optional[Union[BaseModel, str]] = None,
) -> Message:
    """Convert an Anthropic API response into a litemind Message.

    Handles all content block types: text (with citations), tool use/results,
    server tool use (web search, MCP), thinking/redacted thinking, and
    structured output conversion.

    Parameters
    ----------
    anthropic_response : Any
        The raw response object from the Anthropic API.
    response_format : Optional[Union[BaseModel, str]]
        If provided, parses the text response into this Pydantic model.

    Returns
    -------
    Message
        Processed response as a litemind Message with role "assistant".
    """

    # Initialize the processed response:
    processed_response = Message(role="assistant")

    # Check the stop reason and handle special cases:
    stop_reason = anthropic_response.stop_reason
    if stop_reason == "refusal":
        # If the stop reason is refusal, return a message indicating refusal:
        processed_response.append_text("I cannot assist with that request (refusal).")
    elif stop_reason == "max_tokens":
        # Response was truncated due to token limit - add metadata
        processed_response.attributes["truncated"] = True
        processed_response.attributes["stop_reason"] = "max_tokens"
    # "end_turn", "tool_use", "pause_turn", "stop_sequence" are normal completions

    # Text content:
    text_content = ""

    # Variable to hold whether the response is a tool use:
    is_tool_use = False

    # Create a litemind response message from Anthropic's response that includes tool calls:
    for block in anthropic_response.content:
        if block.type == "text":
            # Handle regular text blocks (may include citations)
            text = block.text
            processed_response.append_text(text)
            text_content += text

            # Handle citations if present in the block
            if hasattr(block, "citations") and block.citations:
                # Add citation information as metadata to the last text block
                last_block = processed_response.blocks[-1]
                if not hasattr(last_block, "attributes"):
                    last_block.attributes = {}

                citation_info = []
                for citation in block.citations:
                    if (
                        hasattr(citation, "type")
                        and citation.type == "web_search_result_location"
                    ):
                        citation_info.append(
                            {
                                "type": "web_search_citation",
                                "url": getattr(citation, "url", ""),
                                "title": getattr(citation, "title", ""),
                                "cited_text": getattr(citation, "cited_text", ""),
                            }
                        )

                if citation_info:
                    last_block.attributes["citations"] = citation_info

        elif block.type == "tool_use":
            # Handle regular function tool calls
            processed_response.append_tool_call(
                tool_name=block.name, arguments=block.input, id=block.id
            )
            is_tool_use = True

        elif block.type == "server_tool_use":
            # Handle built-in server tools (web search, MCP, code execution)
            # These are automatically executed by Anthropic, but we track them
            processed_response.append_tool_call(
                tool_name=block.name, arguments=block.input, id=block.id
            )
            is_tool_use = True

        elif block.type == "mcp_tool_use":
            # Handle MCP tool use (similar to server_tool_use)
            processed_response.append_tool_call(
                tool_name=block.name, arguments=block.input, id=block.id
            )
            is_tool_use = True

        elif block.type == "web_search_tool_result":
            # Handle web search results
            search_results: List[WebSearchResultBlock] = block.content

            processed_response.append_tool_use(
                tool_name="web_search",
                arguments={},
                result=search_results,
                id=block.tool_use_id,
            )

        elif block.type == "mcp_tool_result":
            # Handle MCP tool results
            from anthropic.types.beta import BetaTextBlock

            mcp_tool_results: List[BetaTextBlock] = block.content

            # Combine all text blocks into a single string:
            combined_text = "\n".join(
                [result.text for result in mcp_tool_results if result.type == "text"]
            )

            # Append the combined text to the processed response
            processed_response.append_tool_use(
                tool_name="mcp_tool",
                arguments={},
                result=combined_text,
                id=block.tool_use_id,
            )

        elif block.type == "thinking":
            # Handle thinking blocks (existing functionality)
            processed_response.append_thinking(
                block.thinking, signature=block.signature
            )

        elif block.type == "redacted_thinking":
            # Handle redacted thinking blocks (existing functionality)
            processed_response.append_thinking(block.data, redacted=True)

        else:
            raise ValueError(f"Unexpected block type: {block.type}")

    # Handle structured output conversion if requested
    if response_format and not is_tool_use:
        processed_response = json_to_object(
            processed_response, response_format, text_content
        )

    return processed_response



import uuid
from typing import Any, Callable, List, Optional, Type

from numpy import mean
from pydantic import BaseModel

from litemind.agent.augmentations.information.information_base import InformationBase
from litemind.agent.messages.message_block import MessageBlock
from litemind.apis.base_api import BaseApi
from litemind.media.media_base import MediaBase
from litemind.utils.pickle_serialisation import PickleSerializable


class Information(InformationBase, PickleSerializable):
    """
    A piece of information that can be stored in a vector database and retrieved for augmentation.

    Informations can have a hierarchical structure with parent-child relationships.
    This is useful for representing complex informations like PDFs with chapters,
    code repositories with files and folders, or any other nested structure.

    Attributes
    ----------
    media: MediaBase
        The media contained in this Information.
    summary: Optional[str]
        A summary of the information's content.
    metadata: Optional[dict]
        Additional metadata about the information.
    _id: Optional[str]
        A unique identifier for the information.
    embedding: Optional[List[float]]
        Vector embedding of the information content.
    _score: Optional[float]
        A relevance score, usually assigned during retrieval.
    parent: Optional[Information]
        The parent information.
    children: List[Information]
        Child informations. Empty list if no children exist.
    """

    def __init__(
        self,
        media: Optional[MediaBase],
        summary: Optional[str] = None,
        metadata: Optional[dict] = None,
        id: Optional[str] = None,
        score: Optional[float] = None,
        parent: Optional["Information"] = None,
        children: Optional[List["Information"]] = None,
    ):
        """
        Create a new Information.

        An Information must have either ``media`` or ``children``, but
        not both. Leaf informations hold media content directly, while
        parent informations aggregate children.

        Parameters
        ----------
        media : Optional[MediaBase]
            The media content. Must be None if ``children`` is provided.
        summary : Optional[str]
            A text summary of the information's content.
        metadata : Optional[dict]
            Additional metadata (e.g., source, author, date).
        id : Optional[str]
            A unique identifier. Auto-generated (UUID4) if not provided.
        score : Optional[float]
            A relevance score, typically assigned during retrieval.
        parent : Optional[Information]
            The parent information in a hierarchy.
        children : Optional[List[Information]]
            Child informations. Must be None if ``media`` is provided.

        Raises
        ------
        ValueError
            If both ``media`` and ``children`` are None, or both are
            provided.
        """

        # Set attributes:
        # If media is None than children must not be None:
        if media is None and children is None:
            raise ValueError(
                "Cannot create an Information with no media and no children."
            )
        # If media is not None than children must be None:
        if media is not None and children is not None:
            raise ValueError(
                "Cannot create an Information with both media and children."
            )

        # Set the attributes:
        self.media = media
        self.children = children if children is not None else []
        self.summary = summary
        self._metadata = metadata if metadata is not None else {}
        self._id = id or str(uuid.uuid4())
        self._score = score
        self.embedding = None
        self.parent = parent

        # Set the parent for each child information:
        for child in self.children:
            child.parent = self

    def has_type(self, media_type: Type[MediaBase]) -> bool:
        return isinstance(self.media, media_type)

    @property
    def type(self) -> Type[MediaBase]:
        return type(self.media)

    @property
    def content(self) -> Any:
        return self.media.get_content()

    @property
    def metadata(self) -> dict:
        return self._metadata

    @property
    def id(self) -> str:
        return self._id

    @property
    def score(self) -> Optional[float]:
        return self._score

    @score.setter
    def score(self, value: Optional[float]):
        self._score = value

    def set_id(self, value: str) -> None:
        """
        Set the ID of this information.

        Parameters
        ----------
        value: str
            The new ID to set.
        """
        self._id = value

    def add_child(self, document: "Information") -> None:
        """
        Add a child information to this information.

        The child's ``parent`` attribute is set to this information.

        Parameters
        ----------
        document : Information
            The child information to add.
        """
        document.parent = self
        self.children.append(document)

    def compute_embedding(
        self, embedding_function: Optional[Callable[[str], List[float]]] = None
    ) -> List[float]:
        """
        Compute an embedding vector for this information.

        For leaf informations, embeds the media content directly. For
        parent informations with children, recursively computes child
        embeddings and averages them. Cached after first computation.

        Parameters
        ----------
        embedding_function : Optional[Callable[[str], List[float]]]
            A function that takes a string and returns an embedding vector.
            If not provided, uses litemind's default API for embeddings.

        Returns
        -------
        List[float]
            The computed embedding vector.

        Raises
        ------
        ValueError
            If the information has no media and no children.
        """
        if embedding_function is None:
            from litemind import CombinedApi

            api = CombinedApi()

            def embedding_function(text):
                return api.embed_texts([text])[0]

        # If we already have an embedding, just return it
        if self.embedding is not None:
            return self.embedding

        # If we don't have children, compute the embedding for our content
        if not self.children:
            if self.media is None:
                raise ValueError(
                    "Cannot compute embedding: Information has no media and no children."
                )
            self.embedding = embedding_function(self.content)
            return list(self.embedding)

        # If we have children, make sure they all have embeddings
        for child in self.children:
            if child.embedding is None:
                child.compute_embedding(embedding_function)

        # Average the embeddings of our children (and own content if we have media)
        embeddings = [child.embedding for child in self.children]
        if self.media is not None:
            embeddings.append(embedding_function(self.content))
        self.embedding = mean(embeddings, axis=0).tolist()

        # Return the computed embedding
        return self.embedding

    def get_all_informations(self) -> List["Information"]:
        """
        Get all informations in the hierarchy as a flat list.

        Returns
        -------
        List[Information]
            A list of all informations in the hierarchy.
        """
        infos = [self]
        for child in self.children:
            infos.extend(child.get_all_informations())
        return infos

    def summarize(
        self,
        summarization_function: Optional[Callable[[str], str]] = None,
        model: Optional[str] = None,
        api: Optional[BaseApi] = None,
        temperature: float = 0.0,
    ) -> str:
        """
        Generate a text summary for this information.

        Parameters
        ----------
        summarization_function : Optional[Callable[[str], str]]
            A function that takes content and returns a summary string.
            If not provided, uses the LLM API to generate a summary.
        model : Optional[str]
            Model name for LLM-based summarization. Ignored if
            ``summarization_function`` is provided.
        api : Optional[BaseApi]
            API instance for LLM-based summarization. Defaults to
            ``CombinedApi()`` if not provided. Ignored if
            ``summarization_function`` is provided.
        temperature : float
            Sampling temperature for LLM summarization. Ignored if
            ``summarization_function`` is provided.

        Returns
        -------
        str
            The generated summary, also stored in ``self.summary``.

        Raises
        ------
        ValueError
            If ``summarization_function`` is provided together with
            ``api`` or ``model``.
        """
        if summarization_function is None:

            # Use litemind to summarize
            from litemind import CombinedApi
            from litemind.agent.messages.message import Message

            # Create a CombinedApi instance:
            api = api or CombinedApi()

            # Create the system and main messages:
            system_message = Message(
                role="system",
                text="You are a helpful assistant that summarizes informations.",
            )
            main_message = Message(
                role="user",
                text="Please summarize the following document in a concise way.",
            )
            main_message.append_media(self.media)

            # Generate the summary:
            response = api.generate_text(
                messages=[system_message, main_message],
                temperature=temperature,
                model_name=model,
            )

            # Process the response:
            self.summary = response[0].to_plain_text()
        else:
            # Throw an error if api or model are provided:
            if api is not None or model is not None:
                raise ValueError(
                    "Cannot provide api or model if summarization_function is provided."
                )

            self.summary = summarization_function(self.content)

        return self.summary

    def to_message_block(self) -> MessageBlock:
        """
        Convert this information to a MessageBlock for inclusion in a Message.

        Returns
        -------
        MessageBlock
            A message block containing this information's media and metadata.
        """
        return MessageBlock(
            media=self.media,
            metadata=self.metadata,
            id=self.id,
            score=self.score,
        )

    def __copy__(self):
        return Information(
            media=self.media,
            summary=self.summary,
            metadata=self.metadata,
            id=self.id,
            score=self.score,
        )

    def __contains__(self, item):
        return item in self.media

    def __str__(self) -> str:
        return str(self.media)

    def __hash__(self) -> int:
        return hash(self.media)

    def __eq__(self, other) -> bool:
        return self.media == other.media

    def __ne__(self, other) -> bool:
        return not self == other

    def __len__(self) -> int:
        return len(self.media)
